#!/usr/bin/env python3
from dataclasses import dataclass
from pathlib import Path
import argparse
import difflib
import copy
import textwrap
import os
import re


@dataclass
class VersionNumber:
    major: int
    minor: int
    patch: int

    def __str__(self):
        return f"{self.major}.{self.minor}.{self.patch}"

    @property
    def short(self) -> str:
        return f"{self.major}.{self.minor}"


def get_full_filepath(filepath):
    main_directory = Path(os.path.abspath(__file__)).parent.parent
    return Path(main_directory) / filepath


def update_version_number_in_file(relative_filepath, pattern, new_version_number):
    full_filepath = get_full_filepath(relative_filepath)

    with open(full_filepath, "r", encoding="UTF-8") as file:
        file_contents = file.read()
        original = copy.deepcopy(file_contents)

        modified = apply_fixes(pattern, new_version_number, file_contents)
        patch = create_patch(str(full_filepath), original, modified)

        if args.patch_only:
            print(patch)
            return

        if not patch:
            if not args.quiet:
                print("No changes to make to {}".format(full_filepath))
            return

        if not args.quiet:
            print("\n******************************************")
            print("Changes to {}\n".format(full_filepath))
            print(patch)
            print("\n******************************************")

        if args.force:
            make_change = True
        else:
            make_change = yes_or_no("Make changes to {}?".format(full_filepath))

        if make_change:
            with open(full_filepath, "w", encoding="UTF-8") as file:
                file.write(modified)


def bump_version_numbers(
    new_version_number: VersionNumber, next_version_number: VersionNumber
):
    update_version_number_in_file(
        "CITATION.cff", r"^version: (\d+\.\d+\.\d+)", new_version_number
    )
    update_version_number_in_file(
        "manual/sphinx/conf.py", r"^version = \"(\d+\.\d+)\"", new_version_number.short
    )
    update_version_number_in_file(
        "manual/sphinx/conf.py", r"^release = \"(\d+\.\d+\.\d+)\"", new_version_number
    )
    update_version_number_in_file(
        "manual/doxygen/Doxyfile_readthedocs",
        r"^PROJECT_NUMBER\s*=\s*(\d+\.\d+\.\d+)",
        new_version_number,
    )
    update_version_number_in_file(
        "manual/doxygen/Doxyfile",
        r"^PROJECT_NUMBER\s*=\s*(\d+\.\d+\.\d+)",
        new_version_number,
    )
    update_version_number_in_file(
        "CMakeLists.txt",
        r"^set\(_bout_previous_version \"(\d+\.\d+\.\d+)\"\)",
        new_version_number,
    )
    update_version_number_in_file(
        "CMakeLists.txt",
        r"^set\(_bout_next_version \"(\d+\.\d+\.\d+)\"\)",
        next_version_number,
    )
    update_version_number_in_file(
        "tools/pylib/_boutpp_build/backend.py",
        r"_bout_previous_version = \"v(\d+\.\d+\.\d+)\"",
        new_version_number,
    )
    update_version_number_in_file(
        "tools/pylib/_boutpp_build/backend.py",
        r"_bout_next_version = \"v(\d+\.\d+\.\d+)\"",
        next_version_number,
    )


def apply_fixes(pattern, new_version_number, source):
    """Apply the various fixes for each factory to source. Returns
    modified source

    Parameters
    ----------
    pattern
        Regex pattern to apply for replacement
    new_version_number
        New version number to use in replacement
    source
        Text to update
    """

    def get_replacement(match):
        return match[0].replace(match[1], str(new_version_number))

    modified = re.sub(pattern, get_replacement, source, flags=re.MULTILINE)

    return modified


def yes_or_no(question):
    """Convert user input from yes/no variations to True/False"""
    while True:
        reply = input(question + " [y/N] ").lower().strip()
        if not reply or reply[0] == "n":
            return False
        if reply[0] == "y":
            return True


def create_patch(filename, original, modified):
    """Create a unified diff between original and modified"""

    patch = "\n".join(
        difflib.unified_diff(
            original.splitlines(),
            modified.splitlines(),
            fromfile=filename,
            tofile=filename,
            lineterm="",
        )
    )

    return patch


if __name__ == "__main__":
    parser = argparse.ArgumentParser(
        formatter_class=argparse.RawDescriptionHelpFormatter,
        description=textwrap.dedent("""\
            Update the software version number to the specified version, 
            to be given in the form major.minor.patch, 
            e.g. 5.10.3
            
            Where the 3rd ('patch') part of the version is omitted, 
            only the 'major' and 'minor' parts will be used, 
            e.g. 5.10.3 -> 5.10 
            
            For the 'bout-next' version number, 
            the 'minor' version number of the provided version will be incremented by 1, 
            e.g. 5.10.3 -> 5.11.3
            
            """),
    )

    parser.add_argument(
        "--force", "-f", action="store_true", help="Make changes without asking"
    )
    parser.add_argument(
        "--quiet", "-q", action="store_true", help="Don't print patches"
    )
    parser.add_argument(
        "--patch-only", "-p", action="store_true", help="Print the patches and exit"
    )
    parser.add_argument("new_version", help="New (current) version number")
    parser.add_argument(
        "next_version", help="Next version number", nargs="?", default=None
    )

    args = parser.parse_args()

    if args.force and args.patch_only:
        raise ValueError("Incompatible options: --force and --patch")

    new_version = VersionNumber(*map(int, args.new_version.split(".")))
    next_version = (
        VersionNumber(new_version.major, new_version.minor + 1, 0)
        if args.next_version is None
        else VersionNumber(*map(int, args.next_version.split(".")))
    )

    bump_version_numbers(
        new_version_number=new_version, next_version_number=next_version
    )



#! /usr/bin/env python3
# /// script
# requires-python = ">=3.12"
# dependencies = [
#     "pygithub",
# ]
# ///

import argparse
import os
import github


def format_pull(pull: github.PullRequest.PullRequest) -> str:
    """Format a PR as a nice changelog line"""
    url = pull.url.replace("api.", "").replace("pulls", "pull").replace("repos/", "")
    user_url = pull.user.url.replace("api.", "").replace("users/", "")

    return rf"- {pull.title} [\#{pull.number}][{url}] ([{pull.user.login}][{user_url}])"


def pulls_after_tag(repo: github.Repository.Repository, tag: str, release_branch: str):
    """Get a list of PRs merged after a given tag"""

    release = repo.get_release(tag)

    pulls = []

    for pull in repo.get_pulls(
        state="closed", sort="created", base=release_branch, direction="desc"
    ):
        # Unmerged for whatever reason
        if not pull.merged_at:
            continue

        # We're now at PRs merged before the last release
        if pull.merged_at < release.created_at:
            break

        pulls.append(pull)

    return pulls


def pulls_into_RC(repo: github.Repository.Repository, tag: str):
    """Get PRs into the release candidate branch"""
    rc_branch = f"{tag}-rc"
    try:
        repo.get_branch(rc_branch)
    except github.GithubException:
        print(f"No release candidate branch named '{rc_branch}'")
        return []

    return repo.get_pulls(
        state="closed", sort="created", base=rc_branch, direction="desc"
    )


if __name__ == "__main__":
    parser = argparse.ArgumentParser("BOUT++ changelog generator")
    parser.add_argument(
        "--token",
        "-t",
        help="GitHub API token, defaults to '${GITHUB_CHANGELOG_TOKEN}' environment variable",
        default=os.environ.get("GITHUB_CHANGELOG_TOKEN", ""),
    )
    parser.add_argument("last_tag", help="Last git tag")
    parser.add_argument("next_tag", help="Next git tag")
    parser.add_argument("--owner", help="Project organisation", default="boutproject")
    parser.add_argument("--repo", help="Repository name", default="BOUT-dev")
    parser.add_argument(
        "--release-branch",
        help="Release branch name (one of 'master' or 'next')",
        default="next",
    )

    args = parser.parse_args()

    repo_name = f"{args.owner}/{args.repo}"
    repo = github.Github(args.token).get_repo(repo_name)

    print(f"## [{args.next_tag}](https://github.com/{repo_name}/tree/{args.next_tag}")
    print(
        f"\n[Full Changelog](https://github.com/{repo_name}/compare/{args.last_tag}...{args.next_tag})\n"
    )

    pulls = []
    pulls.extend(pulls_into_RC(repo, args.next_tag))
    pulls.extend(pulls_after_tag(repo, args.last_tag, args.release_branch))

    for pull in pulls:
        print(format_pull(pull))



# Generates an input mesh for circular, large aspect-ratio
# simulations:
#
# o Constant magnetic field
# o Curvature output as a 3D logB variable
# o Z is poloidal direction
# o Y is parallel (toroidal)
#
# NOTE: This reverses the standard BOUT/BOUT++ convention
#       so here Bt and Bp are reversed
#

from __future__ import division
from __future__ import print_function
from builtins import range

from numpy import zeros, ndarray, pi, cos, sin, outer, linspace, sqrt

from boututils.datafile import DataFile  # Wrapper around NetCDF4 libraries


def generate(
    nx,
    ny,
    R=2.0,
    r=0.2,  # Major & minor radius
    dr=0.05,  # Radial width of domain
    Bt=1.0,  # Toroidal magnetic field
    q=5.0,  # Safety factor
    mxg=2,
    file="circle.nc",
):
    # q = rBt / RBp
    Bp = r * Bt / (R * q)

    # Minor radius as function of x. Choose so boundary
    # is half-way between grid points

    h = dr / (nx - 2.0 * mxg)  # Grid spacing in r
    rminor = linspace(
        r - 0.5 * dr - (mxg - 0.5) * h, r + 0.5 * dr + (mxg - 0.5) * h, nx
    )

    # mesh spacing in x and y
    dx = ndarray([nx, ny])
    dx[:, :] = r * Bt * h  # NOTE: dx is toroidal flux

    dy = ndarray([nx, ny])
    dy[:, :] = 2.0 * pi / ny

    # LogB = log(1/(1+r/R cos(theta))) =(approx) -(r/R)*cos(theta)
    logB = zeros([nx, ny, 3])  # (constant, n=1 real, n=1 imag)

    # At y = 0, Rmaj = R + r*cos(theta)
    logB[:, 0, 1] = -(rminor / R)

    # Moving in y, phase shift by (toroidal angle) / q
    for y in range(1, ny):
        dtheta = y * 2.0 * pi / ny / q  # Change in poloidal angle

        logB[:, y, 1] = -(rminor / R) * cos(dtheta)
        logB[:, y, 2] = -(rminor / R) * sin(dtheta)

    # Shift angle from one end of y to the other
    ShiftAngle = ndarray([nx])
    ShiftAngle[:] = 2.0 * pi / q

    Rxy = ndarray([nx, ny])
    Rxy[:, :] = r  # NOTE  : opposite to standard BOUT convention

    Btxy = ndarray([nx, ny])
    Btxy[:, :] = Bp

    Bpxy = ndarray([nx, ny])
    Bpxy[:, :] = Bt

    Bxy = ndarray([nx, ny])
    Bxy[:, :] = sqrt(Bt**2 + Bp**2)

    hthe = ndarray([nx, ny])
    hthe[:, :] = R

    print("Writing to file '" + file + "'")

    f = DataFile()
    f.open(file, create=True)

    # Mesh size
    f.write("nx", nx)
    f.write("ny", ny)

    # Mesh spacing
    f.write("dx", dx)
    f.write("dy", dy)

    # Metric components
    f.write("Rxy", Rxy)
    f.write("Btxy", Btxy)
    f.write("Bpxy", Bpxy)
    f.write("Bxy", Bxy)
    f.write("hthe", hthe)

    # Shift
    f.write("ShiftAngle", ShiftAngle)

    # Curvature
    f.write("logB", logB)

    # Input parameters
    f.write("R", R)
    f.write("r", r)
    f.write("dr", dr)
    f.write("Bt", Bt)
    f.write("q", q)
    f.write("mxg", mxg)

    f.close()


def coordinates(
    nx,
    ny,
    nz,
    R=2.0,
    r=0.2,  # Major & minor radius
    dr=0.05,  # Radial width of domain
    Bt=1.0,  # Toroidal magnetic field
    q=5.0,  # Safety factor
    mxg=2,
):
    """
    Returns coordinates (R,Z) as a pair of arrays

    """

    h = dr / (nx - 2.0 * mxg)  # Grid spacing in r
    rminor = linspace(
        r - 0.5 * dr - (mxg - 0.5) * h, r + 0.5 * dr + (mxg - 0.5) * h, nx
    )

    print("Grid spacing: Lx = %e, Lz = %e" % (h, 2.0 * pi * r / nz))

    Rxyz = ndarray([nx, ny, nz])
    Zxyz = ndarray([nx, ny, nz])

    for y in range(0, ny):
        dtheta = y * 2.0 * pi / ny / q  # Change in poloidal angle
        theta = linspace(0, 2.0 * pi, nz, endpoint=False) + dtheta

        Rxyz[:, y, :] = R + outer(rminor, cos(theta))
        Zxyz[:, y, :] = outer(rminor, sin(theta))

    return Rxyz, Zxyz



from __future__ import print_function
from __future__ import division
from past.utils import old_div

#
# Generate the test case using SymPy
#
# Equations are:
#
#
# Delp2(phi) = vort + Sphi

from boutdata.mms import *

####

# from sympy import Integral
# def rms(expr):
#    """ Calculates the RMS value of a function over space
#    """
#    ix = Integral(expr**2, (x, 0, 1))
#    ixy = Integral(ix, (y, 0, 2*pi))
#    ixyz =
#    return sqrt( , (y, 0, 2*pi), (z, 0, 2*pi)).evalf() )


def C(f):
    """Curvature"""
    return bxcvz * diff(f, metric.z)


#######################################
# Inputs, essentially as in BOUT.inp

# Switches
ionvis = False  # Ion Viscosity
Ti = 10  # Ion temperature for viscosity calculation
elecvis = False  # Electron viscosity
resistivity = True

parallel = False  # Parallel dynamics

# Parameters
Tnorm = 5  # Electron Temperature (eV)
Nnorm = 2e18  # Background plasma density (m^-3)
Bnorm = 0.35  # Magnetic field [T]
AA = 2  # Ion atomic mass

#######################################
#

mi_me = AA * 1.67262158e-27 / 9.109e-31
beta_e = qe * Tnorm * Nnorm / (old_div(Bnorm**2, mu0))

# Normalisation parameters
Cs0 = sqrt(qe * Tnorm / (AA * Mp))
Omega_ci = qe * Bnorm / (AA * Mp)
rho_s0 = old_div(Cs0, Omega_ci)

# Define a metric

Rxy = 1.5  # Major radius
Bpxy = 0.35  # Poloidal magnetic field
Bxy = 0.35  # Total magnetic field
Btxy = 0.0  # Toroidal magnetic field
hthe = 1.0  # Poloidal arc length

sinty = 0.0
sbp = 1.0

dx = 2e-5

nx = 130
ZMAX = 1e-3

MXG = 1
MYG = 0

bxcvz = 100  # Curvature

# Normalise

Rxy /= rho_s0
Bpxy /= Bnorm
Btxy /= Bnorm
Bxy /= Bnorm
hthe /= rho_s0

dx /= rho_s0**2 * Bnorm

bxcvz *= rho_s0**2

# Define a metric
metric = Metric()  # Identity

Lx = dx * (nx - 2.0 * MXG)  # Size of the X domain

metric.g11 = (Rxy * Bpxy) ** 2
metric.g22 = old_div(1.0, (hthe**2))
metric.g33 = (sinty**2) * metric.g11 + old_div((Bxy**2), metric.g11)
metric.g12 = 0.0
metric.g13 = -sinty * metric.g11
metric.g23 = -sbp * Btxy / (hthe * Bpxy * Rxy)

metric.J = old_div(hthe, Bpxy)
B = metric.B = Bxy

metric.g_11 = old_div(1.0, metric.g11) + ((sinty * Rxy) ** 2)
metric.g_22 = (Bxy * hthe / Bpxy) ** 2
metric.g_33 = Rxy * Rxy
metric.g_12 = sbp * Btxy * hthe * sinty * Rxy / Bpxy
metric.g_13 = sinty * Rxy * Rxy
metric.g_23 = sbp * Btxy * hthe * Rxy / Bpxy


# Define the manufactured solution in terms of input x,y and z

Ne = 0.9 + 0.9 * x + 0.2 * cos(t) * sin(5.0 * x**2 - 5.0 * z)
Te = 1 + 0.5 * cos(t) * cos(3 * x**2 - 4 * z)
Vort = sin(2 * t) * cos(x - z)
VePsi = 0.5 * cos(7 * t) * cos(3 * x**2 - 3 * z)
Vi = -0.1 * cos(7 * t) * cos(3 * x**2 - 3 * z)

phi = sin(z - x + t) * sin(2.0 * pi * x)  # Must satisfy Dirichlet BCs for now
psi = sin(pi * x) * (
    0.5 * x - cos(7 * t) * sin(3.0 * x**2 - 3 * z)
)  # Must satisfy Dirichlet BCs for now

# Substitute to get in terms of actual x,y,z coordinates

replace = [(x, old_div(metric.x, Lx)), (z, old_div(metric.z, ZMAX))]

Ne = Ne.subs(replace)
Te = Te.subs(replace)
Vort = Vort.subs(replace)
VePsi = VePsi.subs(replace)
Vi = Vi.subs(replace)
phi = phi.subs(replace)
psi = psi.subs(replace)

# Calculate RHS function

Pe = Ne * Te
Ve = VePsi - 0.5 * mi_me * beta_e * psi

Gi = 0.0
Ge = 0.0

nu = 0.0

### Electron density

dNedt = -bracket(phi, Ne, metric) + (old_div(2, B)) * (C(Pe) - Ne * C(phi))

if parallel:
    dNedt -= Ne * Grad_par(Ve, metric) + Vpar_Grad_par(Ve, Ne, metric)

### Electron temperature

dTedt = -bracket(phi, Te, metric) + (old_div(4.0, 3)) * (old_div(Te, B)) * (
    (old_div(7.0, 2)) * C(Te) + (old_div(Te, Ne)) * C(Ne) - C(phi)
)

if parallel:
    dTedt -= Vpar_Grad_par(Ve, Te, metric)
    dTedt += (
        (old_div(2.0, 3.0))
        * Te
        * (
            0.71 * Grad_par(Vi, metric)
            - 1.71 * Grad_par(Ve, metric)
            + 0.71 * (Vi - Ve) * Grad_par(log(Ne), metric)
        )
    )

### Vorticity

dVortdt = -bracket(phi, Vort, metric) + 2.0 * B * C(Pe) / Ne + B * C(Gi) / (3.0 * Ne)

if parallel:
    dVortdt += -Vpar_Grad_par(Vi, Vort, metric) + B ^ 2 * (
        Grad_par(Vi - Ve, metric) + (Vi - Ve) * Grad_par(log(Ne), metric)
    )

### Parallel Ohm's law

if parallel:
    dVePsidt = (
        -bracket(phi, Ve, metric)
        - Vpar_Grad_par(Ve, Ve, metric)
        - mi_me * (old_div(2.0, 3.0)) * Grad_par(Ge, metric)
        - mi_me * nu * (Ve - Vi)
        + mi_me * Grad_par(phi, metric)
        - mi_me * (Te * Grad_par(log(Ne), metric) + 1.71 * Grad_par(Te, metric))
    )
else:
    dVePsidt = 0.0

### Parallel ion velocity

if parallel:
    dVidt = (
        -bracket(phi, Vi, metric)
        - Vpar_Grad_par(Vi, Vi, metric)
        - (old_div(2.0, 3.0)) * Grad_par(Gi, metric)
        - (Grad_par(Te, metric) + Te * Grad_par(log(Ne), metric))
    )
else:
    dVidt = 0.0

#######################

# Create list of all evolving variables
vars = [
    (Ne, dNedt, "Ne"),
    (Te, dTedt, "Te"),
    (Vort, dVortdt, "Vort"),
    (VePsi, dVePsidt, "VePsi"),
    (Vi, dVidt, "Vi"),
]

replace = [(metric.x, x * Lx), (metric.z, z * ZMAX)]

# Loop over variables and print solution, source etc.
for v, dvdt, name in vars:
    # Calculate source
    S = diff(v, t) - dvdt

    dvdx = diff(v, metric.x)
    dvdy = diff(v, metric.y)

    # Substitute back to get in terms of x,y,z
    v = v.subs(replace)
    dvdx = dvdx.subs(replace)
    dvdy = dvdy.subs(replace)
    S = S.subs(replace)

    print("[" + name + "]")
    print("solution = " + exprToStr(v))
    print("\nddx     = " + exprToStr(dvdx))
    print("\nddy     = " + exprToStr(dvdy))
    print("\nsource   = " + exprToStr(S))

# Potential
Sphi = Delp2(phi, metric) - Vort

print("\n[phi]")
print("solution = " + exprToStr(phi.subs(replace)))
print("\nsource   = " + exprToStr(Sphi.subs(replace)))


print("\n\nDelp2 phi = ", Delp2(phi, metric).subs(replace))

##########################################
# Check magnitudes of individual terms

print("\n\nDensity terms:")
print("  bracket : ", (-bracket(phi, Ne, metric)).subs(replace), "\n")
print("  (2/B)*C(Pe) : ", ((old_div(2, B)) * C(Pe)).subs(replace), "\n")
print("  (2/B)*Ne*C(phi) : ", ((old_div(2, B)) * Ne * C(phi)).subs(replace), "\n")


print("\n\nTemperature terms:")
print("  bracket : ", (-bracket(phi, Te, metric)).subs(replace))
print(
    "  C(Te)   : ",
    ((old_div(4.0, 3)) * (old_div(Te, B)) * (old_div(7.0, 2)) * C(Te)).subs(replace),
)
print(
    "  (Te/Ne)*C(Ne) : ",
    ((old_div(4.0, 3)) * (old_div(Te, B)) * (old_div(Te, Ne)) * C(Ne)).subs(replace),
)
print("  C(phi)  : ", (-(old_div(4.0, 3)) * (old_div(Te, B)) * C(phi)).subs(replace))


print("\n\nVorticity terms:")
print("  bracket : ", (-bracket(phi, Vort, metric)).subs(replace))
print("  C(Pe)   : ", (2.0 * B * C(Pe) / Ne).subs(replace))
print("  C(Gi)   : ", B * C(Gi) / (3.0 * Ne))



import sys
import os
import jinja2

from boutconfig import config as conf

from scan_enums import enums


class Field(object):
    """Abstracts over BoutReals and Field2D/3D/Perps

    Provides some helper functions for writing function signatures and
    passing data

    """

    def __init__(self, field_type, dimensions, fdd, boundaries):
        # C++ type of the field, e.g. Field3D
        self.field_type = field_type
        # array: dimensions of the field
        self.dimensions = dimensions
        # Short identifier
        self.fdd = fdd
        self.ddd = fdd
        self.boundaries = boundaries

    def __eq__(self, other):
        try:
            return self.field_type == other.field_type
        except AttributeError:
            return self.field_type == other

    def __ne__(self, other):
        return not (self == other)

    def __str__(self):
        return self.field_type

    @property
    def header(self):
        return self.field_type.lower()

    @property
    def ndims(self):
        return len(self.dimensions)

    def makelist(self, fmt):
        """make an index list based on `fmt` and replace $d and $i"""
        return ", ".join(
            [
                fmt.replace("$d", d).replace("$i", str(i))
                for i, d in enumerate(self.dimensions)
            ]
        )


class Vector(object):
    def __init__(self, vec_type, vdd, field, header):
        self.vec_type = vec_type
        self.vdd = vdd
        self.field = field
        self.fdd = field.fdd
        self.ddd = vdd
        self.header = header

    def __repr__(self):
        return self.vec_type


ops = [("add", "+"), ("mul", "*"), ("truediv", "/"), ("div", "/"), ("sub", "-")]

field2d = Field("Field2D", ["x", "y"], "f2d", ["applyBoundary"])
field3d = Field(
    "Field3D", ["x", "y", "z"], "f3d", ["applyBoundary", "applyParallelBoundary"]
)
fieldperp = Field("FieldPerp", ["x", "z"], "fperp", ["applyBoundary"])

metric = field3d if conf["metric_type"] == "3D" else field2d

vector3d = Vector("Vector3D", "v3d", field3d, "vector3d")
vector2d = Vector("Vector2D", "v2d", metric, "vector2d")
vecs = [vector3d, vector2d]
fields = [field3d, field2d]

if __name__ == "__main__":
    inf = sys.argv[1]
    outf = sys.argv[2]
    tmpf = f"{outf}.tmp"

    with open(tmpf, "w") as out:
        env = jinja2.Environment(loader=jinja2.FileSystemLoader("."), trim_blocks=True)

        template = env.get_template(inf)

        args = dict(
            field3d=field3d,
            field2d=field2d,
            fieldperp=fieldperp,
            ops=ops,
            vecs=vecs,
            fields=fields,
            vector3d=vector3d,
            vector2d=vector2d,
            enums=enums,
            metric_field=metric,
        )
        out.write(template.render(**args) + "\n")

    os.rename(tmpf, outf)



#!/usr/bin/env python
"""
Manually creation of folder structures
To be used with python 2

Dependencies:
-----------
pygraphviz

mmag@fysik.dtu.dk
"""

__authors__ = "Michael Loeiten"
__email__ = "mmag@fysik.dtu.dk"
__version__ = "1.0"
__date__ = "21.01.2016"

import pygraphviz as pgv

# Initialize
tree = pgv.AGraph()

# Appendable lists
files = []
dead_ends = []

# Default node attributes
tree.node_attr["shape"] = "box"
tree.node_attr["style"] = "bold"

# Adding nodes and edges
l0 = "project"
l1 = ["data", "source\nfiles", "driver.py"]
# Append the files
files.append(l1[1])
files.append(l1[2])
# Add the boxes to the mother node
for box in l1:
    tree.add_edge(l0, box)

l2 = ["solver1", "solver2", "BOUT.inp", "run_log.txt"]
# Append the files
files.append(l2[2])
files.append(l2[3])
# Add the boxes to the mother node
for box in l2:
    tree.add_edge("data", box)
tree.add_edge("solver2", "solver2/...")
# Append the dead_end
de = l2[1] + "/..."
dead_ends.append(de)

l3 = ["method1", "method2", "solver1/..."]
for box in l3:
    tree.add_edge("solver1", box)
tree.add_edge("method2", "method2/...")
# Append the dead_end
de = l3[2]
dead_ends.append(de)
de = l3[1] + "/..."
dead_ends.append(de)

l4 = ["nout\ntimestep1", "nout\ntimestep2", "method1/..."]
for box in l4:
    tree.add_edge("method1", box)
tree.add_edge("nout\ntimestep2", "nout\ntimestep2/...")
# Append the dead_end
de = l4[2]
dead_ends.append(de)
de = l4[1] + "/..."
dead_ends.append(de)

l5 = ["mesh1", "mesh2", "nout\ntimestep1/..."]
for box in l5:
    tree.add_edge("nout\ntimestep1", box)
tree.add_edge("mesh2", "mesh2/...")
# Append the dead_end
de = l5[2]
dead_ends.append(de)
de = l5[1] + "/..."
dead_ends.append(de)

l6 = ["additional1", "additional2", "mesh1/..."]
for box in l6:
    tree.add_edge("mesh1", box)
tree.add_edge("additional2", "additional2/...")
# Append the dead_end
de = l6[2]
dead_ends.append(de)
de = l6[1] + "/..."
dead_ends.append(de)

l7 = ["grid_file1", "grid_file2", "additional1/..."]
for box in l7:
    tree.add_edge("additional1", box)
tree.add_edge("grid_file2", "grid_file2/...")
# Append the dead_end
de = l7[2]
dead_ends.append(de)
de = l7[1] + "/..."
dead_ends.append(de)

l8 = [
    "BOUT.inp\n(copy)",
    "BOUT.log",
    "BOUT.dmp",
    "BOUT.restart",
    "(source_files\n(copy))",
    "(grid_file\n(copy))",
]
# Add l8 to the files list
for cur_file in l8:
    files.append(cur_file)
# Append them to the mother node
for box in l8:
    tree.add_edge("grid_file1", box)


# Change colors for the files
for the_file in files:
    member = tree.get_node(the_file)
    #    member.attr['fontcolor'] = 'limegreen'
    member.attr["color"] = "limegreen"

# Change colors for the dead_ends
for dead_end in dead_ends:
    member = tree.get_node(dead_end)
    #    member.attr['fontcolor'] = 'darksalmon'
    member.attr["color"] = "darksalmon"


# Print the graph
print(tree.string())

# Set layout
tree.layout("dot")
# Write to file
tree.draw("folder_tree.svg")



from __future__ import print_function
from builtins import str
from sympy import symbols, sin, diff

###


def Delp2(f):
    """Laplacian in X-Z"""
    d2fdx2 = diff(f, x, 2)
    d2fdz2 = diff(f, z, 2)

    return d2fdx2 + d2fdz2


###

# Define symbols

x = symbols("x")
z = symbols("z")
pi = symbols("pi")

# Define manufactured solution

solution = sin(x * pi)

# Calculate input

input = Delp2(solution)

# Print

print("solution = " + str(solution))
print("input = " + str(input))



from __future__ import division
from builtins import zip
from builtins import range
from past.utils import old_div
import numpy as np
from boututils.datafile import DataFile
from boututils.surface_average import surface_average
from boutdata.collect import collect
from pylab import plot, show, xlabel, ylabel, figure, legend, gca


path = "./data"

gfile = "../cbm18_dens8.grid_nx68ny64.nc"

with DataFile(gfile) as f:
    g = {v: f.read(v) for v in f.keys()}

var = collect("P", path=path)

sol = surface_average(var, g)
# sol=np.mean(var,axis=3)

p0av = collect("P0", path=path)

q = np.zeros(sol.shape)

for i in range(sol.shape[1]):
    q[:, i] = sol[:, i] + p0av[:, 0]


psixy = g.get("psixy")
psi0 = g.get("psi_axis")
psix = g.get("psi_bndry")

xarr = psixy[:, 0]
xarr = old_div((xarr - psi0), (-psi0 + psix))  # for this grid


fig = figure()

nt = q.shape[1]

plot(xarr, p0av, "k", label="t=0")
plot(xarr, q[:, nt / 4], "r", label="t=" + np.str(nt / 4))
plot(xarr, q[:, nt / 2], "b", label="t=" + np.str(nt / 2))
plot(xarr, q[:, 3 * nt / 4], "g", label="t=" + np.str(3 * nt / 4))
plot(xarr, q[:, -1], "k", label="t=" + np.str(nt))

from collections import OrderedDict

handles, labels = gca().get_legend_handles_labels()
by_label = OrderedDict(list(zip(labels, handles)))
legend(list(by_label.values()), list(by_label.keys()))


xlabel(r"$\psi$", fontsize=25)
ylabel(r"$2 \mu_0 <P^2> / B^2$", fontsize=25)

fig.set_tight_layout(True)

show()



#!/usr/bin/env python3

import argparse
import copy
import difflib
import re
import textwrap
from pathlib import Path
from typing import List
from subprocess import run

header_shim_sentinel = "// BOUT++ header shim"

header_warning = f"""\
#pragma once
{header_shim_sentinel}
#warning Header "{{0}}" has moved to "bout/{{0}}". Run `bin/bout-v5-header-upgrader.py` to fix
#include "bout/{{0}}"
"""


def header_needs_moving(header: Path) -> bool:
    """Check if `header` has not yet been moved"""
    with open(header, "r") as f:
        return header_shim_sentinel not in f.read()


def deprecated_header_list(include_path: Path = Path("./include")):
    """List of deprecated header paths (that is, those in bare
    ``include`` directory)

    """
    return include_path.glob("*.hxx")


def write_header_shim(header: Path):
    """Write 'shim' for header, that ``include``s new location"""
    with open(header, "w") as f:
        f.write(header_warning.format(header.name))


def fix_library_header_locations(
    include_path: Path = Path("./include"), quiet: bool = False
):
    unmoved_headers = list(
        filter(header_needs_moving, deprecated_header_list(include_path))
    )
    include_bout_path = include_path / "bout"

    if unmoved_headers == []:
        print("No headers to move!")
        return

    out = run("git diff-index --cached HEAD --quiet", shell=True)
    if out.returncode:
        raise RuntimeError(
            "git index not clean! Please commit or discard any changes before continuing"
        )

    # First we move any existing headers and commit this change, so
    # that history is preserved
    for header in unmoved_headers:
        new_path = include_bout_path / header.name
        if not quiet:
            print(f"Moving '{header}' to '{new_path}'")
        run(f"git mv {header} {new_path}", shell=True, check=True)

    run(r"git commit -m 'Move headers to `include/bout`'", shell=True, check=True)

    # Now we can write the compatibility shim
    for header in unmoved_headers:
        write_header_shim(header)

    run(f"git add {' '.join(map(str, unmoved_headers))}", shell=True, check=True)
    run(
        r"git commit -m 'Add compatibility shims for old header locations'",
        shell=True,
        check=True,
    )


def make_header_regex(deprecated_headers: List[str]) -> re.Pattern:
    """Create a regular expression to match deprecated header locations"""
    deprecated_header_joined = "|".join((header.name for header in deprecated_headers))
    return re.compile(rf'(#include\s+<|")(?:\.\./)?({deprecated_header_joined})(>|")')


def apply_fixes(header_regex, source):
    """Apply all fixes in this module"""
    modified = copy.deepcopy(source)

    return header_regex.sub(r"\1bout/\2\3", modified)


def yes_or_no(question):
    """Convert user input from yes/no variations to True/False"""
    while True:
        reply = input(question + " [y/N] ").lower().strip()
        if not reply or reply[0] == "n":
            return False
        if reply[0] == "y":
            return True


def create_patch(filename, original, modified):
    """Create a unified diff between original and modified"""

    patch = "\n".join(
        difflib.unified_diff(
            original.splitlines(),
            modified.splitlines(),
            fromfile=filename,
            tofile=filename,
            lineterm="",
        )
    )

    return patch


if __name__ == "__main__":
    parser = argparse.ArgumentParser(
        formatter_class=argparse.RawDescriptionHelpFormatter,
        description=textwrap.dedent("""\
            Fix deprecated header locations for BOUT++ v4 -> v5

            All BOUT++ headers are now under ``include/bout`` and
            should be included as ``#include <bout/header.hxx>``. This
            tool will fix such includes.

            For developers: the option ``--move-deprecated-headers``
            will move the headers from ``include`` to
            ``include/bout``, and add a compatibility shim in the old
            location. This option is mutually exclusive with
            ``--files``, and should be used after running this tool
            over the library files.

            WARNING: If any files do need moving, this will create a
            git commit in order to preserve history across file moves.
            If you have staged changes, this tool will not work, so to
            avoid committing undesired or unrelated changes.

            """),
    )

    parser.add_argument(
        "--force", "-f", action="store_true", help="Make changes without asking"
    )
    parser.add_argument(
        "--quiet", "-q", action="store_true", help="Don't print patches"
    )
    parser.add_argument(
        "--patch-only", "-p", action="store_true", help="Print the patches and exit"
    )
    parser.add_argument(
        "--include-path",
        "-i",
        help="Path to `include` directory",
        default="./include",
        type=Path,
    )

    group = parser.add_mutually_exclusive_group(required=True)
    group.add_argument("--files", nargs="*", action="store", help="Input files")
    group.add_argument(
        "--move-deprecated-headers",
        action="store_true",
        help="Move the deprecated headers",
    )

    args = parser.parse_args()

    if args.force and args.patch_only:
        raise ValueError("Incompatible options: --force and --patch")

    deprecated_headers = deprecated_header_list(args.include_path)

    if args.move_deprecated_headers:
        fix_library_header_locations(args.include_path, args.quiet)
        exit(0)

    header_regex = make_header_regex(deprecated_headers)

    for filename in args.files:
        with open(filename, "r") as f:
            contents = f.read()
        original = copy.deepcopy(contents)

        modified = apply_fixes(header_regex, contents)
        patch = create_patch(filename, original, modified)

        if args.patch_only:
            print(patch)
            continue

        if not patch:
            if not args.quiet:
                print(f"No changes to make to {filename}")
            continue

        if not args.quiet:
            print("\n******************************************")
            print(f"Changes to {filename}\n")
            print(patch)
            print("\n******************************************")

        if args.force:
            make_change = True
        else:
            make_change = yes_or_no(f"Make changes to {filename}?")

        if make_change:
            with open(filename, "w") as f:
                f.write(modified)



#!/usr/bin/env python3

import boutpp as bc
import numpy as np

bc.init("-d input -f ../data/BOUT.inp".split(" "))

f = bc.Field3D.fromMesh(None)
f.setAll(np.array([[[1.0]]]))
f2 = bc.Field3D.fromMesh(None)
f2.setAll(np.array([[[2.0]]]))
print(f.getAll())
dump = bc.Datafile()  # get global :(
dump.add(f3d=f, f2d=f2, save_repeat=True)
dump.write()



#!/usr/bin/env python3
#
# Generate manufactured solution and sources for FCI test
#

from math import pi
import warnings

from boututils.boutwarnings import AlwaysWarning
from boutdata.data import BoutOptionsFile
from boutdata.mms import diff, exprToStr, x, y, z
from sympy import sin, cos, sqrt, Expr

warnings.simplefilter("ignore", AlwaysWarning)

f = sin(y - z) + sin(y - 2 * z)
K = cos(z - y)

Lx = 0.1
Ly = 10.0
Lz = 1.0

Bt = 1.0
Bp = 0.05
Bpprime = 0.1

Bpx = Bp + (x - 0.5) * Lx * Bpprime  # Note: x in range [0,1]
B = sqrt(Bpx**2 + Bt**2)


def FCI_grad_par(f: Expr) -> Expr:
    return (Bt * diff(f, y) * 2.0 * pi / Ly + Bpx * diff(f, z) * 2.0 * pi / Lz) / B


def FCI_grad2_par2(f: Expr) -> Expr:
    return FCI_grad_par(FCI_grad_par(f))


def FCI_div_par(f: Expr) -> Expr:
    return Bpx * FCI_grad_par(f / Bpx)


def FCI_div_par_K_grad_par(f: Expr, K: Expr) -> Expr:
    return (K * FCI_grad2_par2(f)) + (FCI_div_par(K) * FCI_grad_par(f))


def FCI_Laplace_par(f: Expr) -> Expr:
    return FCI_div_par(FCI_grad_par(f))


############################################
# Equations solved

options = BoutOptionsFile("data/BOUT.inp")

for name, expr in (
    ("input_field", f),
    ("K", K),
    ("grad_par_solution", FCI_grad_par(f)),
    ("grad2_par2_solution", FCI_grad2_par2(f)),
    ("div_par_solution", FCI_div_par(f)),
    ("div_par_K_grad_par_solution", FCI_div_par_K_grad_par(f, K)),
    ("laplace_par_solution", FCI_Laplace_par(f)),
):
    expr_str = exprToStr(expr)
    print(f"{name} = {expr_str}")
    options[name] = expr_str

options.write("data/BOUT.inp", overwrite=True)



#!/usr/bin/env python3

from sys import argv

from xbout import open_boutdataset

ds = open_boutdataset(
    argv[1] + "/BOUT.dmp.*.nc", keep_yboundaries=1, keep_xboundaries=1
)

ds.bout.animate_list(["rhs", "rhs_check", "error", "f"], animate_over="y", show=True)



#!/usr/bin/env python3


import argparse
import copy
import difflib
import re

try:
    import clang.cindex

    has_clang = True
except ImportError:
    has_clang = False


headers = {"interpolation": {"old": "interpolation.hxx", "new": "interpolation_xz.hxx"}}

interpolations = {
    "Hermite": {"old": "HermiteSpline", "new": "XZHermiteSpline"},
    "Interpolation": {"old": "Interpolation", "new": "XZInterpolation"},
    "MonotonicHermite": {
        "old": "MonotonicHermiteSpline",
        "new": "XZMonotonicHermiteSpline",
    },
    "Bilinear": {"old": "Bilinear", "new": "XZBilinear"},
    "Lagrange4pt": {"old": "Lagrange4pt", "new": "XZLagrange4pt"},
}

factories = {
    "InterpolationFactory": {
        "old": "InterpolationFactory",
        "new": "XZInterpolationFactory",
    }
}


def fix_header_includes(old_header, new_header, source):
    """Replace old_header with new_header in source

    Parameters
    ----------
    old_header: str
        Name of header to be replaced
    new_header: str
        Name of replacement header
    source: str
        Text to search

    """
    return re.sub(
        r"""
        (\s*\#\s*include\s*)     # Preprocessor include
        (<|")
        ({header})              # Header name
        (>|")
        """.format(header=old_header),
        r"\1\2{header}\4".format(header=new_header),
        source,
        flags=re.VERBOSE,
    )


def fix_interpolations(old_interpolation, new_interpolation, source):
    return re.sub(
        r"""
        \b{}\b
        """.format(old_interpolation),
        r"{}".format(new_interpolation),
        source,
        flags=re.VERBOSE,
    )


def clang_parse(filename, source):
    index = clang.cindex.Index.create()
    return index.parse(filename, unsaved_files=[(filename, source)])


def clang_find_interpolations(node, typename, nodes=None):
    if nodes is None:
        nodes = []
    if node.kind == clang.cindex.CursorKind.TYPE_REF:
        if node.type.spelling == typename:
            nodes.append(node)
    for child in node.get_children():
        clang_find_interpolations(child, typename, nodes)
    return nodes


def clang_fix_single_interpolation(
    old_interpolation, new_interpolation, source, location
):
    modified = source
    line = modified[location.line - 1]
    new_line = (
        line[: location.column - 1]
        + new_interpolation
        + line[location.column + len(old_interpolation) - 1 :]
    )
    modified[location.line - 1] = new_line
    return modified


def clang_fix_interpolation(old_interpolation, new_interpolation, node, source):
    nodes = clang_find_interpolations(node, old_interpolation)
    modified = source
    for node in nodes:
        modified = clang_fix_single_interpolation(
            old_interpolation, new_interpolation, modified, node.location
        )
    return modified


def fix_factories(old_factory, new_factory, source):
    return re.sub(
        r"""
        \b{}\b
        """.format(old_factory),
        r"{}".format(new_factory),
        source,
        flags=re.VERBOSE,
    )


def create_patch(filename, original, modified):
    """Create a unified diff between original and modified"""

    patch = "\n".join(
        difflib.unified_diff(
            original.splitlines(),
            modified.splitlines(),
            fromfile=filename,
            tofile=filename,
            lineterm="",
        )
    )

    return patch


def yes_or_no(question):
    """Convert user input from yes/no variations to True/False"""
    while True:
        reply = input(question + " [y/N] ").lower().strip()
        if not reply or reply[0] == "n":
            return False
        if reply[0] == "y":
            return True


def apply_fixes(headers, interpolations, factories, source):
    """Apply all Interpolation fixes to source

    Parameters
    ----------
    headers
        Dictionary of old/new headers
    interpolations
        Dictionary of old/new Interpolation types
    source
        Text to update

    """

    modified = copy.deepcopy(source)

    for header in headers.values():
        modified = fix_header_includes(header["old"], header["new"], modified)
    for interpolation in interpolations.values():
        modified = fix_interpolations(
            interpolation["old"], interpolation["new"], modified
        )
    for factory in factories.values():
        modified = fix_factories(factory["old"], factory["new"], modified)

    return modified


def clang_apply_fixes(headers, interpolations, factories, filename, source):
    # translation unit
    tu = clang_parse(filename, source)

    modified = source

    for header in headers.values():
        modified = fix_header_includes(header["old"], header["new"], modified)

    modified = modified.split("\n")
    for interpolation in interpolations.values():
        modified = clang_fix_interpolation(
            interpolation["old"], interpolation["new"], tu.cursor, modified
        )
    modified = "\n".join(modified)
    for factory in factories.values():
        modified = fix_factories(factory["old"], factory["new"], modified)

    return modified


if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="Fix types of Interpolation objects")

    parser.add_argument("files", action="store", nargs="+", help="Input files")
    parser.add_argument(
        "--force", "-f", action="store_true", help="Make changes without asking"
    )
    parser.add_argument(
        "--quiet", "-q", action="store_true", help="Don't print patches"
    )
    parser.add_argument(
        "--patch-only", "-p", action="store_true", help="Print the patches and exit"
    )
    parser.add_argument(
        "--clang", action="store_true", help="Use libclang if available"
    )

    args = parser.parse_args()

    if args.force and args.patch_only:
        raise ValueError("Incompatible options: --force and --patch")

    if args.clang and not has_clang:
        raise RuntimeError(
            "libclang is not available. Please install libclang Python bindings"
        )

    for filename in args.files:
        with open(filename, "r") as f:
            contents = f.read()
        original = copy.deepcopy(contents)

        if args.clang and has_clang:
            modified = clang_apply_fixes(
                headers, interpolations, factories, filename, contents
            )
        else:
            modified = apply_fixes(headers, interpolations, factories, contents)
        patch = create_patch(filename, original, modified)

        if args.patch_only:
            print(patch)
            continue

        if not patch:
            if not args.quiet:
                print("No changes to make to {}".format(filename))
            continue

        if not args.quiet:
            print("\n******************************************")
            print("Changes to {}\n".format(filename))
            print(patch)
            print("\n******************************************")

        if args.force:
            make_change = True
        else:
            make_change = yes_or_no("Make changes to {}?".format(filename))

        if make_change:
            with open(filename, "w") as f:
                f.write(modified)



from builtins import range
####################################
# computes polslice for time series
####################################

import numpy as np
from boutdata.collect import collect
from boututils.plotpolslice import plotpolslice

###########################
# Specify parameters


path = "./data/"

variable = "P"

p = collect(variable, path=path)

period = 15

grid = "../cbm18_dens8.grid_nx68ny64.nc"

########################################################
# Call plotpolslice once to get extended poloidal grid

r, z, fun = plotpolslice(p[0, :, :, :], grid, period=period, rz=1)

nx = r.shape[0]  # number of points in r
ny = r.shape[1]  # number of points in z
nt = p.shape[0]  # time intervals


fm = np.zeros(
    (nt, nx, ny)
)  # array to store the time sequence of the poloidal cross section

# Compute all time frames

for k in range(nt):
    fm[k, :, :] = plotpolslice(p[k, :, :, :], grid, period=period, rz=0)

np.savez("pslice", fm=fm, z=z, r=r)



#!/usr/bin/env python3

from boutdata import collect
from matplotlib import pyplot
from sys import exit

f = collect("f", path="data", yguards=True, info=False)[1:-1, 1:-1]
sol = collect("sol", path="data", yguards=True, info=False)[1:-1, 1:-1]
error = collect("error", path="data", yguards=True, info=False)[1:-1, 1:-1]
absolute_error = collect("absolute_error", path="data", yguards=True, info=False)[
    1:-1, 1:-1
]

# Note, cells closest to x-boundary in rhs and rhs_check may be slightly different because
# of different way x-boundary cells for D2DXDY are set: in LaplaceXY corner guard cells
# are set so that a 9-point stencil can be used; in the D2DXDY function (used in
# Laplace_perp) first dfdy=DDY(f) is calculated, communicated and has free_o3 x-boundary
# conditions applied, then DDX(dfdy) is returned.
# Therefore here exclude cells closest to the x-boundary so that the difference plotted
# should be small (as controlled by rtol, atol).
rhs = collect("rhs", path="data", yguards=True, info=False)[3:-3, 2:-2]
rhs_check = collect("rhs_check", path="data", yguards=True, info=False)[3:-3, 2:-2]

pyplot.figure()

pyplot.subplot(231)
pyplot.pcolormesh(f)
pyplot.title("f")
pyplot.colorbar()

pyplot.subplot(232)
pyplot.pcolormesh(sol)
pyplot.title("sol")
pyplot.colorbar()

pyplot.subplot(233)
pyplot.pcolormesh(error)
pyplot.title("error")
pyplot.colorbar()

pyplot.subplot(234)
pyplot.pcolormesh(absolute_error)
pyplot.title("absolute_error")
pyplot.colorbar()

pyplot.subplot(235)
pyplot.pcolormesh(rhs)
pyplot.title("rhs")
pyplot.colorbar()

pyplot.subplot(236)
pyplot.pcolormesh(rhs - rhs_check)
pyplot.title("rhs diff")
pyplot.colorbar()

pyplot.show()

exit(0)



import zoidberg as zb
import numpy as np
import sys
import boutconfig as bc


def rotating_ellipse(
    nx=68,
    ny=16,
    nz=128,
    npoints=421,
    xcentre=5.5,
    I_coil=0.01,
    curvilinear=True,
    rectangular=False,
    fname="rotating-ellipse.fci.nc",
    a=0.4,
    Btor=2.5,
):
    yperiod = 2 * np.pi / 5.0
    field = zb.field.RotatingEllipse(
        xcentre=xcentre,
        I_coil=I_coil,
        radius=2 * a,
        yperiod=yperiod,
        Btor=Btor,
    )
    # Define the y locations
    ycoords = np.linspace(0.0, yperiod, ny, endpoint=False)

    if rectangular:
        print("Making rectangular poloidal grid")
        poloidal_grid = zb.poloidal_grid.RectangularPoloidalGrid(
            nx, nz, 1.0, 1.0, Rcentre=xcentre
        )
    elif curvilinear:
        print("Making curvilinear poloidal grid")
        inner = zb.rzline.shaped_line(
            R0=xcentre, a=a / 2.0, elong=0, triang=0.0, indent=0, n=npoints
        )
        outer = zb.rzline.shaped_line(
            R0=xcentre, a=a, elong=0, triang=0.0, indent=0, n=npoints
        )

        print("creating grid...")
        poloidal_grid = zb.poloidal_grid.grid_elliptic(inner, outer, nx, nz)

    # Create the 3D grid by putting together 2D poloidal grids
    grid = zb.grid.Grid(poloidal_grid, ycoords, yperiod, yperiodic=True)
    maps = zb.make_maps(grid, field, quiet=True)
    zb.write_maps(grid, field, maps, str(fname), metric2d=bc.isMetric2D())


if __name__ == "__main__":
    rotating_ellipse(fname=sys.argv[1])



import numpy as np
from scipy import optimize
import matplotlib.pyplot as plt

Te_ref = np.loadtxt("temperature.csv", delimiter=",")
Te_ref[:, 0] *= 1e-4  # Convert X axis to m


def te_function(ypos, mid, wwid, w0, w1, w2, Tmax, Tmin, clip=False):
    width = w0 + ((ypos - mid) * w1 + (ypos - mid) ** 2 * w2) * np.exp(
        -(((ypos - mid) / wwid) ** 2)
    )

    if clip:
        width = np.clip(width, 1e-10, None)

    return Tmax - 0.5 * (1 + np.tanh((ypos - mid) / width)) * (Tmax - Tmin)


popt, pcov = optimize.curve_fit(
    te_function,
    Te_ref[:, 0],
    Te_ref[:, 1],
    p0=[2.2e-4, 1e-4, 1e-4, 0.0, 0.0, 0.960, 0.190],
)

print(popt)

xfit = np.linspace(Te_ref[0, 0], Te_ref[-1, 0], 100)

plt.plot(xfit, te_function(xfit, *popt, clip=True), "-k")
plt.plot(Te_ref[:, 0], Te_ref[:, 1], "or")
plt.show()



from __future__ import print_function
from __future__ import division
from builtins import range
from past.utils import old_div

# the Scientific Python netCDF 3 interface
# http://dirac.cnrs-orleans.fr/ScientificPython/
# from Scientific.IO.NetCDF import NetCDFFile as Dataset
# the 'classic' version of the netCDF4 python interface
# http://code.google.com/p/netcdf4-python/
import numpy as np
from netCDF4 import Dataset
from numpy import dtype  # array module from http://numpy.scipy.org

"""
This example writes some surface pressure and temperatures
The companion program sfc_pres_temp_rd.py shows how to read the netCDF
data file created by this program.

This example demonstrates the netCDF Python API.
It will work either with the Scientific Python NetCDF version 3 interface
(http://dirac.cnrs-orleans.fr/ScientificPython/)
of the 'classic' version of the netCDF4 interface. 
(http://netcdf4-python.googlecode.com/svn/trunk/docs/netCDF4_classic-module.html)
To switch from one to another, just comment/uncomment the appropriate
import statements at the beginning of this file.

Jeff Whitaker <jeffrey.s.whitaker@noaa.gov> 20070202
"""
# Adapted for BOUT++ by
# George Breyiannis, JAEA, Nov 2013
#

# the output array to write will be nx x ny
ny = 100
nx = ny + 4

# dy of grid
dy = old_div(1.0, np.float(ny))
dx = dy
# create grid
dxarr = np.zeros((nx, ny), dtype="float32") + dx
dyarr = np.zeros((nx, ny), dtype="float32") + dy

xarr = np.arange(0.0, np.float(nx), 1.0, dtype="float32") * dx
yarr = np.arange(0.0, np.float(ny), 1.0, dtype="float32") * dy

# compute initial variables

rho = np.zeros((nx, ny), dtype="float32") + old_div(25.0, (36.0 * np.pi))
p = np.zeros((nx, ny), dtype="float32") + old_div(5.0, (12.0 * np.pi))

rho = 1.0
p = old_div(rho, 3.0)

v_x = np.zeros((nx, ny), dtype="float32")
Bx = np.zeros((nx, ny), dtype="float32")

for y in range(ny):
    v_x[:, y] = -np.sin(2.0 * np.pi * yarr[y])
    Bx[:, y] = -np.sin(2.0 * np.pi * yarr[y])

# Bx=Bx/np.sqrt(4.*np.pi)


v_y = np.zeros((nx, ny), dtype="float32")
By = np.zeros((nx, ny), dtype="float32")

for x in range(nx):
    v_y[x, :] = np.sin(2.0 * np.pi * xarr[x])
    By[x, :] = np.sin(4.0 * np.pi * xarr[x])

# By=By/np.sqrt(4.*np.pi)


# Domain inside core (periodic)

ixseps1 = nx
ixseps2 = nx

# open a new netCDF file for writing.
ncfile = Dataset("otv.grd.128.nc", "w", format="NETCDF3_CLASSIC")

# output data.

# create the nx and ny dimensions.
ncfile.createDimension("x", nx)
ncfile.createDimension("y", ny)
ncfile.createDimension("single", 1)

# create and write nx,ny variables
nxx = ncfile.createVariable("nx", "i4", ("single"))
nyy = ncfile.createVariable("ny", "i4", ("single"))

nxx[:] = nx
nyy[:] = ny

# Define the coordinate variables. They will hold the coordinate
# information, that is, xarr,yarr
dx = ncfile.createVariable("dx", dtype("float32").char, ("x", "y"))
dy = ncfile.createVariable(
    "dy",
    dtype("float32").char,
    (
        "x",
        "y",
    ),
)

# write data to coordinate vars.
dx[:, :] = dxarr
dy[:, :] = dyarr

# create and write ixseps* dimensions.
ix1 = ncfile.createVariable("ixseps1", "i4", ("single"))
ix2 = ncfile.createVariable("ixseps2", "i4", ("single"))

ix1[:] = ixseps1
ix2[:] = ixseps2

# create the corresponding variables
rho0 = ncfile.createVariable("rho0", dtype("float32").char, ("x", "y"))
p0 = ncfile.createVariable("p0", dtype("float32").char, ("x", "y"))
v0_x = ncfile.createVariable("v0_x", dtype("float32").char, ("x", "y"))
v0_y = ncfile.createVariable("v0_y", dtype("float32").char, ("x", "y"))
B0x = ncfile.createVariable("B0x", dtype("float32").char, ("x", "y"))
B0y = ncfile.createVariable("B0y", dtype("float32").char, ("x", "y"))

# write data to variables.

rho0[:, :] = rho
p0[:, :] = p
v0_x[:, :] = v_x
v0_y[:, :] = v_y
B0x[:, :] = Bx
B0y[:, :] = By


ncfile.close()
print("*** SUCCESS writing file otv.grd.py.nc!")



class Enum(object):
    def __init__(self, name, vals, source):
        self.name = name.strip()
        self.vals = [x.strip() for x in vals.split(",")]
        if "None" in self.vals:
            self.vals.remove("None")
        self.lower = self.name.lower()
        self.extra = {}
        self.source = source

    def __str__(self):
        return self.name


enums = {}
# Repeat the filename as source so we don't have to do any fancy logic
# to prepend "bout/" in front of one of the files
for inf, source in (
    ("../../../include/bout/bout_types.hxx", "bout/bout_types.hxx"),
    ("other_enums.hxx", "other_enums.hxx"),
):
    with open(inf) as f:
        for line in f:
            line = line.strip()
            if line.startswith("enum"):
                assert line.startswith("enum class ")
                what = line[5 + 6 :]
                if __name__ == "__main__":
                    print(what, inf)
                name, vals = what.split("{")
                assert vals.endswith("};")
                enum = Enum(name, vals[:-2], source)
                enums[enum.name] = enum
            if line.startswith("constexpr"):
                (enum, str), (val,) = [x.split() for x in line[10:].split("=")]
                val = val[:-1]  # remove ;
                val = val.split("::")[-1]
                if __name__ == "__main__":
                    print(enum, str, val)

enums["DIFF_METHOD"].extra = {
    "DIFF_DEFAULT": "deflt",
    "DEFAULT": "deflt",
    "DIFF_U1": "u1",
    "U1": "u1",
    "DIFF_U2": "u2",
    "U2": "u2",
    "DIFF_C2": "c2",
    "C2": "c2",
    "DIFF_W2": "w2",
    "W2": "w2",
    "DIFF_W3": "w3",
    "W3": "w3",
    "DIFF_C4": "c4",
    "C4": "c4",
    "DIFF_U3": "u3",
    "U3": "u3",
    "DIFF_FFT": "fft",
    "FFT": "fft",
    "DIFF_SPLIT": "split",
    "SPLIT": "split",
    "DIFF_S2": "s2",
    "S2": "s2",
}
enums["REGION"].extra = {
    "RGN_ALL": "all",
    "RGN_NOBNDRY": "nobndry",
    "RGN_NOX": "nox",
    "RGN_NOY": "noy",
    "RGN_NOZ": "noz",
}
enums["BRACKET_METHOD"].extra = {
    "BRACKET_STD": "standard",
    "STD": "standard",
    "BRACKET_SIMPLE": "simple",
    "SIMPLE": "simple",
    "BRACKET_ARAKAWA": "arakawa",
    "ARAKAWA": "arakawa",
    "BRACKET_CTU": "ctu",
    "CTU": "ctu",
}
enums["CELL_LOC"].extra = {
    "CELL_DEFAULT": "deflt",
    "DEFAULT": "deflt",
    "CELL_CENTRE": "centre",
    "CENTRE": "centre",
    "CELL_CENTER": "centre",
    "CENTER": "centre",
    "CELL_XLOW": "xlow",
    "XLOW": "xlow",
    "CELL_YLOW": "ylow",
    "YLOW": "ylow",
    "CELL_ZLOW": "zlow",
    "ZLOW": "zlow",
    "CELL_VSHIFT": "vshift",
    "VSHIFT": "vshift",
}

enums = list(enums.values())



from setuptools import Extension, setup
from Cython.Build import cythonize
import os
import numpy

from boutconfig import config as conf

libs = []
ldirs = []
lextra = []
for l in conf["libs"].split():
    if l.startswith("-l"):
        libs.append(l[2:])
    elif l.startswith("-L"):
        ldirs.append(l[2:])
    else:
        lextra.append(l)

include = []
flags = []
for i in conf["cflags"].split():
    if i.startswith("-I"):
        include.append(i[2:])
    else:
        flags.append(i)
include.append(numpy.get_include())

extensions = [
    Extension(
        "boutpp",
        ["boutpp.pyx"],
        include_dirs=include,
        libraries=libs,
        library_dirs=ldirs,
        extra_compile_args=flags,
        extra_link_args=lextra,
    )
]

# Setting LDSHARED sets the linker. However, setting it might drop
# flags - thus it is not set by default. If the compiler used here is
# different to the python compiler, LDSHARED might need to be set. In
# that case it might be required to check the linking commands, and
# add the apropriate flags.

setup(
    ext_modules=cythonize(
        extensions,
        language_level=3,
        compiler_directives=dict(binding=True, embedsignature=True),
    )
)



#! /usr/bin/env python

# MPark.Variant
#
# This script uploads a directory to Wandbox (http://melpon.org/wandbox),
# which is an online compiler environment, and prints a permalink to the
# uploaded code. We use this to provide a "Try it online" version of the
# library to make the barrier to entry as low as possible.
#
# This script was adapted from the script proposed in
# https://github.com/melpon/wandbox/issues/153.
#
# To know how to use this script: ./wandbox.py --help
#
# Copyright Louis Dionne 2015
#
# Distributed under the Boost Software License, Version 1.0.
# (See accompanying file LICENSE.md or copy at http://boost.org/LICENSE_1_0.txt)
#
# Copyright Michael Park, 2017
#
# Distributed under the Boost Software License, Version 1.0.
# (See accompanying file LICENSE.md or copy at http://boost.org/LICENSE_1_0.txt)

import argparse
import fnmatch
import json
import os
import re
import urllib2

# Post the given JSON data to Wandbox's API, and return the result
# as a JSON object.
def upload(options):
    request = urllib2.Request('http://melpon.org/wandbox/api/compile.json')
    request.add_header('Content-Type', 'application/json')
    response = urllib2.urlopen(request, json.dumps(options))
    return json.loads(response.read())

# Returns a list of the '.hpp' headers in the given directory and in
# subdirectories.
#
# The path must be absolute, and the returned paths are all absolute too.
def headers(path):
    return [
        os.path.join(dir, file)
            for (dir, _, files) in os.walk(path)
                for file in fnmatch.filter(files, "*.hpp")
    ]

def main():
    parser = argparse.ArgumentParser(description=
        """Upload a directory to Wandbox (http://melpon.org/wandbox).

           On success, the program prints a permalink to the uploaded
           directory on Wandbox and returns 0. On error, it prints the
           response from the Wandbox API and returns 1.

           Note that the comments are stripped from all the headers in the
           uploaded directory.
        """
    )
    parser.add_argument('directory', type=str, help=
        """A directory to upload to Wandbox.

           The path may be either absolute or relative to the current directory.
           However, the names of the files uploaded to Wandbox will all be
           relative to this directory. This way, one can easily specify the
           directory to be '/some/project/include', and the uploaded files
           will be uploaded as-if they were rooted at '/some/project/include'
        """)
    parser.add_argument('main', type=str, help=
        """The main source file.

           The path may be either absolute or relative to the current directory.
        """
    )
    args = parser.parse_args()
    directory = os.path.abspath(args.directory)
    if not os.path.exists(directory):
        raise Exception("'%s' is not a valid directory" % args.directory)

    cpp = os.path.abspath(args.main)
    if not os.path.exists(cpp):
        raise Exception("'%s' is not a valid file name" % args.main)

    response = upload({
        'code': open(cpp).read().strip(),
        'codes': [{
            'file': os.path.relpath(header, directory).replace('\\', '/'),
            'code': open(header).read().strip()
        } for header in headers(directory)],
        'options': 'warning,optimize,c++14',
        'compiler': 'clang-7.0.0',
        'save': True,
        'compiler-option-raw': '-I.'
    })

    if response['status'] == '0':
        print response['url']
        return 0
    else:
        print response
        return 1

exit(main())



from distutils.core import setup
from Cython.Build import cythonize
import os

os.environ["CXX"] = "mpic++"
os.environ["CC"] = "mpic++"

setup(
    ext_modules=cythonize(
        "debug.pyx",
        language="c++",  # generate C++ code
    )
)



#!/usr/bin/env python3

import argparse
import copy
import difflib
import itertools
import textwrap
import warnings

from boutdata.data import BoutOptionsFile, BoutOptions
from boututils.boutwarnings import AlwaysWarning


def case_sensitive_init(self, name="root", parent=None):
    self._sections = dict()
    self._keys = dict()
    self._name = name
    self._parent = parent
    self.comments = dict()
    self.inline_comments = dict()
    self._comment_whitespace = dict()


# Monky-patch BoutOptions to make sure it's case sensitive
BoutOptions.__init__ = case_sensitive_init


# This should be a list of dicts, each containing "old", "new" and optionally "new_values".
# The values of "old"/"new" keys should be the old/new names of input file values or
# sections. The value of "new_values" is a dict containing replacements for values of the
# option. "old_type" optionally specifies the type of the old value of the option; for
# example this is needed for special handling of boolean values.
REPLACEMENTS = [
    {"old": "mesh:paralleltransform", "new": "mesh:paralleltransform:type"},
    {"old": "fci", "new": "mesh:paralleltransform"},
    {"old": "interpolation", "new": "mesh:paralleltransform:xzinterpolation"},
    {
        "old": "fft:fft_measure",
        "new": "fft:fft_measurement_flag",
        "old_type": bool,
        "new_values": {False: "estimate", True: "measure"},
    },
    {"old": "TIMESTEP", "new": "timestep"},
    {"old": "NOUT", "new": "nout"},
    {"old": "ddx", "new": "mesh:ddx"},
    {"old": "ddy", "new": "mesh:ddy"},
    {"old": "ddz", "new": "mesh:ddz"},
    {"old": "laplace:laplace_nonuniform", "new": "laplace:nonuniform"},
    {"old": "mesh:dump_format", "new": "dump_format"},
    {"old": "solver:ATOL", "new": "solver:atol"},
    {"old": "solver:RTOL", "new": "solver:rtol"},
    # This was inconsistent in the library
    {"old": "All", "new": "all"},
    # The following haven't been changed, but are frequently spelt with the wrong case
    {"old": "mxg", "new": "MXG"},
    {"old": "myg", "new": "MYG"},
    {"old": "nxpe", "new": "NXPE"},
    {"old": "nype", "new": "NYPE"},
    {"old": "mesh:NX", "new": "mesh:nx"},
    {"old": "mesh:NY", "new": "mesh:ny"},
    {"old": "mesh:shiftangle", "new": "mesh:ShiftAngle"},
    {"old": "mesh:shiftAngle", "new": "mesh:ShiftAngle"},
    {"old": "mesh:zshift", "new": "mesh:zShift"},
    {"old": "mesh:StaggerGrids", "new": "mesh:staggergrids"},
    {"old": "output:shiftOutput", "new": "output:shiftoutput"},
    {"old": "output:ShiftOutput", "new": "output:shiftoutput"},
    {"old": "output:shiftInput", "new": "output:shiftinput"},
    {"old": "output:ShiftInput", "new": "output:shiftinput"},
    {"old": "output:flushFrequency", "new": "output:flushfrequency"},
    {"old": "output:FlushFrequency", "new": "output:flushfrequency"},
    {"old": "TwistShift", "new": "twistshift"},
    {"old": "zmin", "new": "ZMIN"},
    {"old": "zmax", "new": "ZMAX"},
    {"old": "ZPERIOD", "new": "zperiod"},
    # 'restart' can be either a section or a value, so move all the
    # section:values instead
    {"old": "restart:parallel", "new": "restart_files:parallel"},
    {"old": "restart:flush", "new": "restart_files:flush"},
    {"old": "restart:guards", "new": "restart_files:guards"},
    {"old": "restart:floats", "new": "restart_files:floats"},
    {"old": "restart:openclose", "new": "restart_files:openclose"},
    {"old": "restart:enabled", "new": "restart_files:enabled"},
    {"old": "restart:init_missing", "new": "restart_files:init_missing"},
    {"old": "restart:shiftOutput", "new": "restart_files:shiftOutput"},
    {"old": "restart:shiftInput", "new": "restart_files:shiftInput"},
    {"old": "restart:flushFrequency", "new": "restart_files:flushFrequency"},
]

for section, derivative in itertools.product(
    ["ddx", "ddy", "ddz", "diff"], ["First", "Second", "Fourth", "Flux", "Upwind"]
):
    REPLACEMENTS.append(
        {
            "old": f"mesh:{section}:{derivative}",
            "new": f"mesh:{section}:{derivative.lower()}",
        }
    )

DELETED = ["dump_format"]

for section, value in itertools.product(
    ["output", "restart"],
    [
        "floats",
        # Following are not yet implemented in OptionsNetCDF. Not yet
        # clear if they need to be, or can be safely removed
        # "shiftoutput",
        # "shiftinput",
        # "flushfrequency",
        # "parallel",
        # "guards",
        # "openclose",
        # "init_missing",
    ],
):
    DELETED.append(f"{section}:{value}")


def parse_bool(bool_expression):
    try:
        bool_expression_lower = bool_expression.lower()
    except AttributeError:
        # bool_expression was not a str: no need to lower
        bool_expression_lower = bool_expression

    if bool_expression_lower in ["true", "y", "t", 1, True]:
        return True
    elif bool_expression_lower in ["false", "n", "f", 0, False]:
        return False
    else:
        raise RuntimeError(
            f"Expected boolean option. Could not parse {bool_expression}"
        )


def already_fixed(replacement, options_file):
    """Check if the options_file already has already had this particular fix applied"""
    # The old key is there and the new one isn't, then it's definitely not fixed
    if replacement["old"] in options_file and replacement["new"] not in options_file:
        return False
    # If the new isn't there, there's nothing to fix
    if replacement["new"] not in options_file:
        return True
    # If we don't need to fix values, we're done
    if "new_values" not in replacement:
        return True
    # Check if the current value is acceptable
    return options_file[replacement["new"]] in replacement["new_values"].values()


def fix_replacements(replacements, options_file):
    """Change the names of options in options_file according to the list
    of dicts replacements

    """
    for replacement in replacements:
        try:
            if already_fixed(replacement, options_file):
                continue
            options_file.rename(replacement["old"], replacement["new"])
        except KeyError:
            pass
        except TypeError as e:
            raise RuntimeError(
                "Could not apply transformation: '{old}' -> '{new}' to file '{0}', due to error:"
                "\n\t{1}".format(options_file.filename, e.args[0], **replacement)
            ) from e
        else:
            if "old_type" in replacement:
                # Special handling for certain types, replicating what BOUT++ does
                if replacement["old_type"] is bool:
                    # The original value must be something that BOUT++ recognises as a
                    # bool.
                    # replacement["new_values"] must contain both True and False keys.
                    old_value = parse_bool(options_file[replacement["new"]])
                    options_file[replacement["new"]] = replacement["new_values"][
                        old_value
                    ]
                else:
                    raise ValueError(
                        f"Error in REPLACEMENTS: type {replacement['type']} is not handled"
                    )
            else:
                # Option values are just a string
                if "new_values" in replacement:
                    old_value = options_file[replacement["new"]]
                    try:
                        old_value = old_value.lower()
                    except AttributeError:
                        # old_value was not a str, so no need to convert to lower case
                        pass

                    try:
                        options_file[replacement["new"]] = replacement["new_values"][
                            old_value
                        ]
                    except KeyError:
                        # No replacement given for this value: keep the old one
                        pass


def remove_deleted(deleted, options_file):
    """Remove each key that appears in 'deleted' from 'options_file'"""

    for key in deleted:
        # Better would be options_file.pop(key, None), but there's a
        # bug in current implementation
        if key in options_file:
            del options_file[key]


def apply_fixes(replacements, deleted, options_file):
    """Apply all fixes in this module"""

    modified = copy.deepcopy(options_file)

    fix_replacements(replacements, modified)

    remove_deleted(deleted, modified)

    return modified


def yes_or_no(question):
    """Convert user input from yes/no variations to True/False"""
    while True:
        reply = input(question + " [y/N] ").lower().strip()
        if not reply or reply[0] == "n":
            return False
        if reply[0] == "y":
            return True


def create_patch(filename, original, modified):
    """Create a unified diff between original and modified"""

    patch = "\n".join(
        difflib.unified_diff(
            original.splitlines(),
            modified.splitlines(),
            fromfile=filename,
            tofile=filename,
            lineterm="",
        )
    )

    return patch


def possibly_apply_patch(patch, options_file, quiet=False, force=False):
    """Possibly apply patch to options_file. If force is True, applies the
    patch without asking, overwriting any existing file. Otherwise,
    ask for confirmation from stdin

    """
    if not quiet:
        print("\n******************************************")
        print("Changes to {}\n".format(options_file.filename))
        print(patch)
        print("\n******************************************")

    if force:
        make_change = True
    else:
        make_change = yes_or_no("Make changes to {}?".format(options_file.filename))
    if make_change:
        options_file.write(overwrite=True)
    return make_change


if __name__ == "__main__":
    parser = argparse.ArgumentParser(
        formatter_class=argparse.RawDescriptionHelpFormatter,
        description=textwrap.dedent("""\
            Fix input files for BOUT++ v5+

            Please note that this will only fix input options in sections with
            standard or default names. You may also need to fix options in custom
            sections.

            Warning! Even with no fixes, there may still be changes as this script
            will "canonicalise" the input files:

              * nested sections are moved to be under their parent section, while
                preserving relative order

              * empty sections are removed

              * floating point numbers may have their format changed, although the
                value will not change

              * consecutive blank lines will be reduced to a single blank line

              * whitespace around equals signs will be changed to exactly one space

              * trailing whitespace will be removed

              * comments will always use '#'

            Files that change in this way will have the "canonicalisation" patch
            presented first. If you choose not to apply this patch, the "upgrade
            fixer" patch will still include it."""),
    )

    parser.add_argument("files", action="store", nargs="+", help="Input files")

    force_patch_group = parser.add_mutually_exclusive_group()
    force_patch_group.add_argument(
        "--force", "-f", action="store_true", help="Make changes without asking"
    )
    force_patch_group.add_argument(
        "--patch-only", "-p", action="store_true", help="Print the patches and exit"
    )

    parser.add_argument(
        "--quiet", "-q", action="store_true", help="Don't print patches"
    )
    parser.add_argument(
        "--accept-canonical",
        "-c",
        action="store_true",
        help="Automatically accept the canonical patch",
    )
    parser.add_argument(
        "--canonical-only",
        "-k",
        action="store_true",
        help="Only check/fix canonicalisation",
    )

    args = parser.parse_args()

    warnings.simplefilter("ignore", AlwaysWarning)

    for filename in args.files:
        with open(filename, "r") as f:
            original_source = f.read()

        try:
            original = BoutOptionsFile(filename)
        except ValueError:
            pass

        canonicalised_patch = create_patch(filename, original_source, str(original))
        if canonicalised_patch and not args.patch_only:
            print(
                "WARNING: original input file '{}' not in canonical form!".format(
                    filename
                )
            )
            applied_patch = possibly_apply_patch(
                canonicalised_patch,
                original,
                args.quiet,
                args.force or args.accept_canonical,
            )
            # Re-read input file
            if applied_patch:
                original_source = str(original)

        if args.canonical_only:
            continue

        try:
            modified = apply_fixes(REPLACEMENTS, DELETED, original)
        except RuntimeError as e:
            print(e)
            continue
        patch = create_patch(filename, original_source, str(modified))

        if args.patch_only:
            print(patch)
            continue

        if not patch:
            if not args.quiet:
                print("No changes to make to {}".format(filename))
            continue

        possibly_apply_patch(patch, modified, args.quiet, args.force)



#!/usr/bin/env python3

import os
import re
from boututils.run_wrapper import shell


class Requirements(object):
    """Checks if test cases meet their specified requirements. Test
    scripts can contain a line starting with #requires or #Requires,
    optionally followed by a colon, then a Python expression which
    combines boolean values.

    Examples:

    #requires netcdf

    #requires not make

    #requires not (boutpp and netcdf)

    The individual requirements (netcdf, make, etc.)
    are gathered from the bout-config scipt, or from executable
    scripts in the "requirements" directory (where this code is).

    """

    def __init__(self, path=None, verbose=False):
        selflocation = os.path.realpath(__file__)
        selflocation = selflocation.rsplit("/", 1)[0]

        if path is None:
            path = selflocation

        # Get list of files in subdirectory, excluding common temporaries,
        # hidden files, and python .py and .pyc files
        requirements_list = [
            x
            for x in os.listdir(path)
            if not (("#" in x) or ("~" in x) or (x[0] == ".") or (".py" in x))
        ]

        if verbose:
            print("======= Requirement checks ========")

        self._verbose = verbose
        self._requirements = {}
        for requirement in requirements_list:
            status, out = shell(os.path.join(path, requirement), pipe=True)
            self.add(requirement, (status == 0))

        with open(selflocation + "/../../bin/bout-config") as configfile:
            config = configfile.read()
            matches = re.findall("^has_(.*)=(.*)", config, re.MULTILINE)
            for match in matches:
                key = match[0]
                value = match[1]
                yesno = {'"yes"': True, '"no"': False}
                try:
                    value = yesno[value]
                except KeyError:
                    print(
                        "Error parsing " + match + ' - %s is not "yes"/"no"' % match[1]
                    )
                else:
                    self.add(key, value)

        # Now have dictionary of requirements, true/false.

    def check(self, script):
        """Tests if the requirements of the given script (file name) are met.
        Returns the result (True/False) and the requirements expression.
        """

        with open(script, "r", encoding="utf-8") as filein:
            contents = filein.read()

            # Find all lines starting with '#requires' or '#Requires:'
            match = re.findall("^\s*\#\s?[Rr]equires:?(.*)", contents, re.MULTILINE)
            # Iterate over all expressions to evaluate
            for expr in match:
                try:
                    ret = eval(expr, self._requirements)
                except:
                    print("Parsing '%s' from file %s failed." % (expr, script))
                    print("A list of known variable follows:")
                    for key, value in self._requirements.items():
                        if "_" not in key:
                            print("  %s => %s" % (key, value))
                    raise
                else:
                    # Only return if this one causes the test to skip
                    if ret is False:
                        return ret, expr

        return True, ""  # Default, if no requirements specified

    def add(self, requirement, value, override=False):
        """Add a requirement value to the dictionary. If the value is
        already defined with another value, it will raise an
        exception, unless override is set to True."""
        if requirement in self._requirements:
            old = self._requirements[requirement]
            if old != value and override is False:
                raise RuntimeError(
                    "Overwriting %s with %s in key %s" % (value, old, requirement)
                )
        self._requirements[requirement] = value
        if self._verbose:
            print("{0} => {1}".format(requirement, value))

    def has(self, requirement):
        """Check if a requrirement is available"""
        return self._requirements[requirement]



import sys
import re
import platform

def main():
    output = sys.stdin.read()

    print(output)

    print("-" * 50)

    time = int(re.search(r"\d+ tests? from \d+ test suites? ran. \((\d+) ms total\)", output).group(1))

    dwarf4 = any(["DWARF4" in arg for arg in sys.argv[1:]])
    dwarf5 = any(["DWARF5" in arg for arg in sys.argv[1:]])
    clang = any(["clang" in arg for arg in sys.argv[1:]])
    # Somehow -gdwarf-4 clang is fast after a completely unrelated PR? Weird. Something to do with static linking...?
    # https://github.com/jeremy-rifkin/cpptrace/pull/22
    expect_slow = False

    if platform.system() == "Windows":
        # For some reason SymInitialize is super slow on github's windows runner and it alone takes 250ms. Nothing we
        # can do about that.
        threshold = 350 # ms
    else:
        threshold = 100 # ms

    if expect_slow:
        if time > threshold:
            print(f"Success (expecting slow): Test program took {time} ms")
        else:
            print(f"Error (expecting slow): Test program took {time} ms")
            sys.exit(1)
    else:
        if time > threshold:
            print(f"Error: Test program took {time} ms")
            sys.exit(1)
        else:
            print(f"Success: Test program took {time} ms")


main()



from __future__ import division
from __future__ import print_function

from boutdata.mms import Metric, sin, Div_par, Grad_par, exprToStr, diff, y, t
from math import pi

# Length of the y domain
Ly = 10.0

# metric tensor
metric = Metric()  # Identity

# Define solution in terms of input x,y,z

f = 1 + 0.1 * sin(2 * y - t)
k = 1 + 0.1 * sin(y)

# Turn solution into real x and z coordinates
replace = [(y, metric.y * 2 * pi / Ly)]

f = f.subs(replace)
k = k.subs(replace)

##############################
# Calculate time derivatives

dfdt = Div_par(k * Grad_par(f))

#############################
# Calculate sources

Sf = diff(f, t) - dfdt

# Substitute back to get input y coordinates
replace = [(metric.y, y * Ly / (2 * pi))]

k = k.subs(replace)
f = f.subs(replace)

Sf = Sf.subs(replace)

print("[mesh]")
print("k = " + exprToStr(k))

print("\n[f]")
print("solution = " + exprToStr(f))
print("\nsource = " + exprToStr(Sf))



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# BOUT++ documentation build configuration file, created by
# sphinx-quickstart on Mon Jan 23 17:21:15 2017.
#
# This file is execfile()d with the current directory set to its
# containing dir.
#
# Note that not all possible configuration values are present in this
# autogenerated file.
#
# All configuration values have a default; values that are commented out
# serve to show the default.

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
import argparse

try:
    from breathe import apidoc

    has_breathe = True
except ImportError:
    print("breathe module not installed")
    has_breathe = False

import os
import subprocess
import sys

sys.path.append("../../tools/pylib")

# Are we running on readthedocs?
on_readthedocs = os.environ.get("READTHEDOCS") == "True"

if on_readthedocs:
    from unittest.mock import MagicMock

    class Mock(MagicMock):
        __all__ = [
            "foo",
        ]

        @classmethod
        def __getattr__(cls, name):
            return MagicMock()

    MOCK_MODULES = [
        "netCDF4",
        "mayavi2",
        "enthought",
        "enthought.mayavi",
        "enthought.mayavi.scripts",
        "enthought.tvtk",
        "enthought.tvtk.api",
        "scipy",
        "scipy.ndimage",
        "scipy.interpolate",
        "scipy.integrate",
        "tvtk",
        "tvtk.tools",
        "tvtk.api",
        "scipy.ndimage.filters",
        "scipy.ndimage.morphology",
        "scipy.spatial",
        "past",
        "crosslines",
    ]
    sys.modules.update((mod_name, Mock()) for mod_name in MOCK_MODULES)
    print(os.environ)
    print(sys.argv)
    os.system("which clang-format")
    os.system("which clang-format-6.0")
    subprocess.run("git submodule update --init --recursive", shell=True)
    cmake = (
        "cmake -S ../.. "
        " -B bout_build"
        " -DBOUT_USE_FFTW=ON"
        " -DBOUT_USE_LAPACK=OFF"
        " -DBOUT_ENABLE_PYTHON=ON"
        " -DBOUT_UPDATE_GIT_SUBMODULE=OFF"
        " -DBOUT_TESTS=OFF"
        " -DBOUT_ALLOW_INSOURCE_BUILD=ON"
        f" -DPython3_ROOT_DIR={sys.exec_prefix}"
    )
    subprocess.run(f"echo {cmake}", shell=True)
    subprocess.run(f"{cmake}", shell=True, check=True)
    subprocess.run("cmake --build bout_build", shell=True, check=True)

    # Add the build directory to sys.path so that sphinx picks up the built
    # Python modules
    sys.path.append("bout_build/tools/pylib")


if has_breathe:
    # Run doxygen to generate the XML sources
    if on_readthedocs:
        subprocess.call("cd ../doxygen; doxygen Doxyfile_readthedocs", shell=True)
    else:
        subprocess.call("cd ../doxygen; doxygen Doxyfile", shell=True)
    # Now use breathe.apidoc to autogen rst files for each XML file
    apidoc_args = argparse.Namespace(
        destdir="_breathe_autogen/",
        dryrun=False,
        force=True,
        notoc=False,
        outtypes=("file"),
        project="BOUT++",
        rootpath="../doxygen/bout/xml",
        suffix="rst",
        members=True,
        quiet=False,
    )
    apidoc_args.rootpath = os.path.abspath(apidoc_args.rootpath)
    if not os.path.isdir(apidoc_args.destdir):
        if not apidoc_args.dryrun:
            os.makedirs(apidoc_args.destdir)
    apidoc.recurse_tree(apidoc_args)
    for key, value in apidoc.TYPEDICT.items():
        apidoc.create_modules_toc_file(key, value, apidoc_args)

    # -- Options for breathe extension ----------------------------------------

    breathe_projects = {"BOUT++": "../doxygen/bout/xml"}
    breathe_default_project = "BOUT++"
    breathe_default_members = ("members",)

# -- General configuration ------------------------------------------------

# If your documentation needs a minimal Sphinx version, state it here.
#
# needs_sphinx = '1.0'

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    "sphinx.ext.coverage",
    "sphinx.ext.mathjax",
    "sphinx.ext.autodoc",
    "sphinx.ext.napoleon",
    "sphinx.ext.todo",
]

if has_breathe:
    extensions.append("breathe")

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# How to parse markdown files
try:
    from recommonmark.parser import CommonMarkParser

    source_parsers = {".md": CommonMarkParser}
except ImportError:
    print("recommonmark module not installed")
    source_parsers = {}

# The suffix(es) of source filenames.
# You can specify multiple suffix as a list of string:
#
source_suffix = [".rst", ".md"]

# The master toctree document.
master_doc = "index"

# General information about the project.
project = "BOUT++"
author = "B. Dudson and The BOUT++ team"
copyright = f"2017-2023, {author}"

# The version info for the project you're documenting, acts as replacement for
# |version| and |release|, also used in various other places throughout the
# built documents.
#
# The short X.Y version.
version = "5.2"
# The full version, including alpha/beta/rc tags.
release = "5.2.0"

# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
#
# This is also used if you do content translation via gettext catalogs.
# Usually you set "language" from the command line for these cases.
language = "en"

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This patterns also effect to html_static_path and html_extra_path
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = "sphinx"

# If true, `todo` and `todoList` produce output, else they produce nothing.
todo_include_todos = False

# Tell sphinx what the primary language being documented is.
primary_domain = "cpp"

# Tell sphinx what the pygments highlight language should be.
highlight_language = "cpp"

# Turn on figure numbering
numfig = True

# The default role for text marked up `like this`
default_role = "any"

# Handle multiple parameters on one line correctly (in Python docs)
napoleon_use_param = False

# -- Options for HTML output ----------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = "sphinx_book_theme"
# Theme options are theme-specific and customize the look and feel of a theme
# further.  For a list of options available for each theme, see the
# documentation.
#
html_theme_options = dict(
    repository_url="https://github.com/boutproject/BOUT-dev",
    repository_branch="master",
    path_to_docs="manual/sphinx",
    use_edit_page_button=True,
    use_repository_button=True,
    use_issues_button=True,
    home_page_in_toc=False,
)


# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ["_static"]


def setup(app):
    app.add_css_file("css/custom.css")


# -- Options for HTMLHelp output ------------------------------------------

# Output file base name for HTML help builder.
htmlhelp_basename = "BOUTdoc"


# -- Options for LaTeX output ---------------------------------------------

latex_elements = {
    # The paper size ('letterpaper' or 'a4paper').
    #
    # 'papersize': 'letterpaper',
    # The font size ('10pt', '11pt' or '12pt').
    #
    # 'pointsize': '10pt',
    # Additional stuff for the LaTeX preamble.
    #
    "preamble": r"""
\usepackage[utf8]{inputenc}
\usepackage[T1]{fontenc}
\makeatletter
\def\UTFviii@defined#1{%
  \ifx#1\relax
      ?%
  \else\expandafter
    #1%
  \fi
}
\makeatother


""",
    # Latex figure (float) alignment
    #
    # 'figure_align': 'htbp',
}

# Grouping the document tree into LaTeX files. List of tuples
# (source start file, target name, title,
#  author, documentclass [howto, manual, or own class]).
latex_documents = [
    (master_doc, "BOUT.tex", "BOUT++ Documentation", "The BOUT++ team", "manual"),
]


# -- Options for manual page output ---------------------------------------

# One entry per manual page. List of tuples
# (source start file, name, description, authors, manual section).
man_pages = [(master_doc, "bout", "BOUT++ Documentation", [author], 1)]


# -- Options for Texinfo output -------------------------------------------

# Grouping the document tree into Texinfo files. List of tuples
# (source start file, target name, title, author,
#  dir menu entry, description, category)
texinfo_documents = [
    (
        master_doc,
        "BOUT",
        "BOUT++ Documentation",
        author,
        "BOUT",
        "One line description of project.",
        "Miscellaneous",
    ),
]



from __future__ import print_function
from numpy import *
from boutdata.collect import collect

path = "./data/"
data = collect("P", path=path)

print("Saving P..")
fa = fft.fft(data, axis=3)
save("fp", rollaxis(fa, 0, 4))



from __future__ import print_function

#
# Generate the test case using SymPy
#
# Equations are:
#
# df/dt = DDY(g)
# dg/dt = DDY(f)
#
#

from sympy import cos, sin, diff

from boutdata.mms import exprToStr, y, t

# Define the manufactured solution

f = y + cos(y) - sin(t) * cos(0.5 * y)
g = y**2 + sin(y) + cos(t) * cos(0.1 * y * y)

# Calculate gradients in x for boundaries

dfdy = diff(f, y)
dgdy = diff(g, y)

# Calculate sources

Sf = diff(f, t) - dgdy
Sg = diff(g, t) - dfdy

#######################

print("\n[f]")
print("solution = " + exprToStr(f))
print("ddy      = " + exprToStr(dfdy))
print("source   = " + exprToStr(Sf))

print("\n[g]")
print("solution = " + exprToStr(g))
print("ddy      = " + exprToStr(dgdy))
print("source   = " + exprToStr(Sg))



# Generate MMS solutions for tokamak geometry

from __future__ import division
from __future__ import print_function

from boutdata.mms import *

from sympy import sin, cos

from math import pi

shape = SimpleTokamak()
metric = shape.metric()  # Get the metric tensor

###
# Define solution in normalised x,y coordinates
# NOTE: These are orthogonal tokamak coordinates
#       so y is poloidal angle, not parallel coordinate

phi = (sin(z - x + t) + 0.001 * cos(y - z)) * sin(
    2.0 * pi * x
)  # Must satisfy Dirichlet BCs for now

Psi = 1e-2 * cos(4 * x**2 + z - y)  # + sin(t)*sin(3*x + 2*z - y))

U = 2.0 * cos(2 * t) * cos(x - z + 4 * y)

P = 1 + 0.5 * cos(t) * cos(3 * x**2 - z + y) + 0.005 * sin(y - z) * sin(t)

P0 = 2 + cos(x * pi)  # Pressure pedestal
J0 = 1 - x - sin(x * pi) ** 2 * cos(y)  # Parallel current
bxcvz = -((1.0 / shape.Rxy) ** 2) * cos(y)  # Curvature

eta = 1e-1  # core_resist =  1 / core_lund
hyperresist = -1e-6  # negative -> none

viscos_par = 1.0

ZMAX = 1

nonlinear = True
diamag = True

# Turn solution into real x and z coordinates
# NOTE: Z is shifted, so y is now the parallel coordinate
zShift = shape.zShift.subs(x, metric.x / metric.Lx)
sinty = shape.sinty.subs(x, metric.x / metric.Lx)
replace = [(x, metric.x / metric.Lx), (z, metric.z / ZMAX + zShift)]

phi = phi.subs(replace)
Psi = Psi.subs(replace)
U = U.subs(replace)
P = P.subs(replace)

P0 = P0.subs(replace)
J0 = J0.subs(replace)

# Calculate time derivatives

B0 = metric.B


def Grad_parP(f):
    result = Grad_par(f, metric)
    if nonlinear:
        result -= B0 * bracket(Psi, f, metric)

    return result


##########################################
# Normalise

MU0 = 4.0e-7 * pi
Mi = 2.0 * 1.6726e-27  # Ion mass [kg]

Bbar = 1  # Tesla
Lbar = 1  # meters
# Note: With this choice, all geometrical normalisation factors go to 1.

Jpar = Delp2(Psi, metric)

##########################################
# Parallel electric field

dPsidt = -Grad_parP(phi) + eta * Jpar

if hyperresist > 0.0:
    dPsidt -= eta * hyperresist * Delp2(Jpar, metric)

# a = Grad_par(B0*phi, metric)/B0
# b = bracket(Psi, B0*phi, metric)
# c = eta * Jpar
# d = eta*hyperresist * Delp2(Jpar, metric)

##########################################
# Vorticity

dUdt = (
    B0**2 * b0xGrad_dot_Grad(Psi, J0, metric)
    + bxcvz * diff(P, metric.z)
    - (B0**2) * Grad_parP(Jpar)
)
if nonlinear:
    # Bracket method '0' (STD) goes to b0xGrad_dot_Grad
    dUdt -= b0xGrad_dot_Grad(phi, U, metric)

# a = B0**2 * b0xGrad_dot_Grad(Psi, J0, metric)
# b = bxcvz*diff(P, metric.z)
# c = (B0**2) * Grad_parP(Jpar)

##########################################
# Pressure

dPdt = -b0xGrad_dot_Grad(phi, P0, metric)

if nonlinear:
    # Bracket method '0' (STD) goes to b0xGrad_dot_Grad
    dPdt -= b0xGrad_dot_Grad(phi, P, metric)

vars = [(Psi, dPsidt, "Psi"), (U, dUdt, "U"), (P, dPdt, "P")]

# Change back to normalised coordinates
# NOTE: Z remains shifted, so y is still parallel coordinate
#       Need to scale the toroidal angle part of z, not zShift part
replace = [
    (metric.x, x * metric.Lx),
    (metric.z, (z - shape.zShift) * ZMAX + shape.zShift),
]

# For applying boundary conditions to shifted fields, remove zShift
replace_shiftbc = [(metric.x, x * metric.Lx), (metric.z, (z - shape.zShift) * ZMAX)]

# print "MAG: ", exprMag(a.subs(replace)), exprMag(b.subs(replace)), exprMag(c.subs(replace))#, exprMag(d.subs(replace))


# Potential
if diamag:
    # Delp2(phi + 0.5*P/B0) = U + Sphi
    Sphi = Delp2(phi + 0.5 * P / B0, metric) - U
else:
    # Delp2(phi) = U + Sphi
    Sphi = Delp2(phi, metric) - U
phi = phi.subs(replace)
Sphi = Sphi.subs(replace)
print("[phi]")
print("solution = " + exprToStr(phi))
print("\nsource = " + exprToStr(Sphi))

Jpar = Jpar.subs(replace)
print("\n[J]")
print("solution = " + exprToStr(Jpar))


# Loop over variables and print solution, source etc.
for f, dfdt, name in vars:
    # Calculate source
    S = diff(f, t) - dfdt

    # Differentials for boundary conditions
    dfdx = diff(f, metric.x)
    dfdy = diff(f, metric.y)

    # Substitute back to get in terms of x,y,z

    fbc = f.subs(replace_shiftbc)
    f = f.subs(replace)
    dfdt = dfdt.subs(replace)
    S = S.subs(replace)
    dfdx = dfdx.subs(replace)
    dfdy = dfdy.subs(replace)

    print("\n[" + name + "]")
    print("solution = " + exprToStr(f))
    print("solution_zshift = " + exprToStr(fbc))
    print("\nddx = " + exprToStr(dfdx))
    print("\nddy = " + exprToStr(dfdy))
    print("\nsource = " + exprToStr(S))

    # print("\nSource magnitude: %e" % exprMag(S))



from __future__ import print_function
from builtins import str

#
# Generate the test case using SymPy
#
# Equations are:
#
# d/dt(n) = -[phi,n] + alpha*(phi - n) - kappa*DDZ(phi) + Sn
# d/dt(vort) = -[phi,vort] + alpha*(phi - n)  + Svort
#
# Delp2(phi) = vort + Sphi

from sympy import symbols, cos, sin, diff

####


def bracket(f, g):
    """
    Calculates [f,g] symbolically
    """

    dfdx = diff(f, x)
    dfdz = diff(f, z)

    dgdx = diff(g, x)
    dgdz = diff(g, z)

    return dfdz * dgdx - dfdx * dgdz


def DDZ(f):
    return diff(f, z)


def Delp2(f):
    """Laplacian in X-Z"""
    d2fdx2 = diff(f, x, 2)
    d2fdz2 = diff(f, z, 2)

    return d2fdx2 + d2fdz2


def Delp4(f):
    d4fdx4 = diff(f, x, 4)
    d4fdz4 = diff(f, z, 4)

    return d4fdx4 + d4fdz4


def exprToStr(expr):
    """Convert a sympy expression to a string for BOUT++ input"""
    return str(expr).replace("**", "^")  # Replace exponent operator


####

# Parameters
alpha = 1.0
kappa = 0.5
Dn = 1.0
Dvort = 1.0

# Define symbols

x = symbols("x")
z = symbols("z")
t = symbols("t")
pi = symbols("pi")

# Define the manufactured solution

n = 0.9 + 0.9 * x + 0.2 * cos(10 * t) * sin(5.0 * x**2 - 2 * z)
vort = 0.9 + 0.7 * x + 0.2 * cos(7 * t) * sin(2.0 * x**2 - 3 * z)
phi = sin(pi * x) * (
    0.5 * x - cos(7 * t) * sin(3.0 * x**2 - 3 * z)
)  # Must satisfy Dirichlet BCs for now

# Calculate gradients in x for boundaries

dndx = diff(n, x)
dvortdx = diff(vort, x)
dphidx = diff(phi, x)

# Calculate RHS function

dndt = -bracket(phi, n) + alpha * (phi - n) - kappa * DDZ(phi) + Dn * Delp2(n)

dvortdt = -bracket(phi, vort) + alpha * (phi - n) + Dvort * Delp2(vort)

# Calculate sources

Sn = diff(n, t) - dndt
Svort = diff(vort, t) - dvortdt

Sphi = Delp2(phi) - vort

#######################

print("[n]")
print("solution = " + exprToStr(n))
print("\nddx     = " + exprToStr(dndx))
print("\nsource   = " + exprToStr(Sn))

print("\n[vort]")
print("solution = " + exprToStr(vort))
print("\nddx     = " + exprToStr(dvortdx))
print("\nsource   = " + exprToStr(Svort))

print("\n[phi]")
print("solution = " + exprToStr(phi))
print("\nddx     = " + exprToStr(dphidx))
print("\nsource   = " + exprToStr(Sphi))



#!/usr/bin/env python3

import argparse
import copy
import difflib
import re

# Dictionary of factory methods that may need updating
factories = {
    "Interpolation": {
        "factory_name": "InterpolationFactory",
        "type_name": "Interpolation",
        "create_method": "create",
    },
    "InvertPar": {
        "factory_name": "InvertPar",
        "type_name": "InvertPar",
        "create_method": "create",
        "old_create_method": "Create",
        "arguments_changed": True,
    },
    "Mesh": {"factory_name": "Mesh", "type_name": "Mesh", "create_method": "Create"},
    "Laplacian": {
        "factory_name": "Laplacian",
        "type_name": "Laplacian",
        "create_method": "create",
    },
    "LaplaceXZ": {
        "factory_name": "LaplaceXZ",
        "type_name": "LaplaceXZ",
        "create_method": "create",
    },
    "SolverFactory": {
        "factory_name": "SolverFactory",
        "type_name": "Solver",
        "create_method": "createSolver",
    },
    "Solver": {
        "factory_name": "Solver",
        "type_name": "Solver",
        "create_method": "create",
    },
}


def find_factory_calls(factory, source):
    """Find all the places where the factory creation method is called,
    and return a list of the variable names

    Parameters
    ----------
    factory
        Dictionary containing 'factory_name' and 'create_method'
    source
        Text to search

    """
    return re.findall(
        r"""
        \s*([\w_]+)                  # variable
        \s*=\s*
        {factory_name}::
        .*{create_method}.*
        """.format(**factory),
        source,
        re.VERBOSE,
    )


def find_type_pointers(factory, source):
    return re.findall(
        r"""
        \b{type_name}\s*\*\s*   # Type name and pointer
        ([\w_]+)\s*;            # Variable name
        """.format(**factory),
        source,
        re.VERBOSE,
    )


def fix_declarations(factory, variables, source):
    """Fix the declaration of varables in source. Returns modified source

    Replaces `Type*` with either `std::unique_ptr<Type>` for
    declarations, or with `auto` for initialisations.

    Parameters
    ----------
    factory
        Dictionary of factory information
    variables
        List of variable names
    source
        Text to update

    """

    for variable in variables:
        # Declarations
        source = re.sub(
            r"""
            (.*?)(class\s*)?         # optional "class" keyword
            \b({type_name})\s*\*\s*  # Type-pointer
            ({variable_name})\s*;    # Variable
            """.format(type_name=factory["type_name"], variable_name=variable),
            r"\1std::unique_ptr<\3> \4{nullptr};",
            source,
            flags=re.VERBOSE,
        )

        # Declarations with initialisation from factory
        source = re.sub(
            r"""
            (.*?)(class\s*)?       # optional "class" keyword
            ({type_name})\s*\*\s*  # Type-pointer
            ({variable_name})\s*   # Variable
            =\s*                   # Assignment from factory
            ({factory_name}::.*{create_method}.*);
            """.format(variable_name=variable, **factory),
            r"\1auto \4 = \5;",
            source,
            flags=re.VERBOSE,
        )

        # Declarations with zero initialisation
        source = re.sub(
            r"""
            (.*?)(?:class\s*)?     # optional "class" keyword
            ({type_name})\s*\*\s*  # Type-pointer
            ({variable_name})\s*   # Variable
            =\s*                   # Assignment
            (0|nullptr|NULL);
            """.format(variable_name=variable, **factory),
            r"\1std::unique_ptr<\2> \3{nullptr};",
            source,
            flags=re.VERBOSE,
        )

    return source


def fix_deletions(variables, source):
    """Remove `delete` statements of variables. Returns modified source

    Parameters
    ----------
    variables
        List of variable names
    source
        Text to update

    """

    for variable in variables:
        source = re.sub(
            r"(.*;?)\s*(delete\s*{variable})\s*;".format(variable=variable),
            r"\1",
            source,
        )

    return source


def fix_create_method(factory, source):
    """Fix change of name of factory `create` method"""

    if "old_create_method" not in factory:
        return source
    old_create_pattern = re.compile(
        r"({factory_name})\s*::\s*{old_create_method}\b".format(**factory)
    )
    if not old_create_pattern.findall(source):
        return source

    if factory.get("arguments_changed", False):
        print(
            "**WARNING** Arguments of {factory_name}::{create_method} have changed, and your current arguments may not work."
            " Please consult the documentation for the new arguments.".format(**factory)
        )

    return re.sub(
        r"({factory_name})\s*::\s*{old_create_method}\b".format(**factory),
        r"\1::{create_method}".format(**factory),
        source,
    )


def yes_or_no(question):
    """Convert user input from yes/no variations to True/False"""
    while True:
        reply = input(question + " [y/N] ").lower().strip()
        if not reply or reply[0] == "n":
            return False
        if reply[0] == "y":
            return True


def apply_fixes(factories, source, all_declarations=False):
    """Apply the various fixes for each factory to source. Returns
    modified source

    Parameters
    ----------
    factories
        Dictionary of factory properties
    source
        Text to update
    """

    modified = source

    for factory in factories.values():
        modified = fix_create_method(factory, modified)
        variables = find_factory_calls(factory, modified)
        if all_declarations:
            variables = variables + find_type_pointers(factory, modified)
        modified = fix_declarations(factory, variables, modified)
        modified = fix_deletions(variables, modified)

    return modified


def create_patch(filename, original, modified):
    """Create a unified diff between original and modified"""

    patch = "\n".join(
        difflib.unified_diff(
            original.splitlines(),
            modified.splitlines(),
            fromfile=filename,
            tofile=filename,
            lineterm="",
        )
    )

    return patch


if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="Fix types of factory-created objects")

    parser.add_argument("files", action="store", nargs="+", help="Input files")
    parser.add_argument(
        "--force", "-f", action="store_true", help="Make changes without asking"
    )
    parser.add_argument(
        "--quiet", "-q", action="store_true", help="Don't print patches"
    )
    parser.add_argument(
        "--patch-only", "-p", action="store_true", help="Print the patches and exit"
    )
    parser.add_argument(
        "--all-declarations",
        "-a",
        action="store_true",
        help="Fix all declarations of factory types, not just variables created from factories",
    )

    args = parser.parse_args()

    if args.force and args.patch_only:
        raise ValueError("Incompatible options: --force and --patch")

    for filename in args.files:
        with open(filename, "r") as f:
            contents = f.read()
        original = copy.deepcopy(contents)

        modified = apply_fixes(
            factories, contents, all_declarations=args.all_declarations
        )
        patch = create_patch(filename, original, modified)

        if args.patch_only:
            print(patch)
            continue

        if not patch:
            if not args.quiet:
                print("No changes to make to {}".format(filename))
            continue

        if not args.quiet:
            print("\n******************************************")
            print("Changes to {}\n".format(filename))
            print(patch)
            print("\n******************************************")

        if args.force:
            make_change = True
        else:
            make_change = yes_or_no("Make changes to {}?".format(filename))

        if make_change:
            with open(filename, "w") as f:
                f.write(modified)



#!/usr/bin/env python3

from boutdata import collect
from boututils.datafile import DataFile
from matplotlib import pyplot
import numpy
from sys import exit

# Set default colormap
pyplot.rcParams["image.cmap"] = "viridis"
# pyplot.rcParams['image.cmap'] = 'plasma'

g11 = collect("g11", path="data", xguards=False, yguards=False)
g12 = collect("g12", path="data", xguards=False, yguards=False)
g13 = collect("g13", path="data", xguards=False, yguards=False)
g22 = collect("g22", path="data", xguards=False, yguards=False)
g23 = collect("g23", path="data", xguards=False, yguards=False)
g33 = collect("g33", path="data", xguards=False, yguards=False)
g_11 = collect("g_11", path="data", xguards=False, yguards=False)
g_12 = collect("g_12", path="data", xguards=False, yguards=False)
g_13 = collect("g_13", path="data", xguards=False, yguards=False)
g_22 = collect("g_22", path="data", xguards=False, yguards=False)
g_23 = collect("g_23", path="data", xguards=False, yguards=False)
g_33 = collect("g_33", path="data", xguards=False, yguards=False)

grid = DataFile("data/d3d_119919.nc")
R = grid["Rxy"][2:-2, :]
Z = grid["Zxy"][2:-2, :]

pyplot.subplot(231)
pyplot.pcolor(R, Z, g11)
pyplot.colorbar()
pyplot.title("g11")

pyplot.subplot(232)
pyplot.pcolor(R, Z, g12)
pyplot.colorbar()
pyplot.title("g12")

pyplot.subplot(233)
pyplot.pcolor(R, Z, g13)
pyplot.colorbar()
pyplot.title("g13")

pyplot.subplot(234)
pyplot.pcolor(R, Z, g22)
pyplot.colorbar()
pyplot.title("g22")

pyplot.subplot(235)
pyplot.pcolor(R, Z, g23)
pyplot.colorbar()
pyplot.title("g23")

pyplot.subplot(236)
pyplot.pcolor(R, Z, g33)
pyplot.colorbar()
pyplot.title("g33")


pyplot.figure()

pyplot.subplot(231)
pyplot.pcolor(R, Z, g_11)
pyplot.colorbar()
pyplot.title("g_11")

pyplot.subplot(232)
pyplot.pcolor(R, Z, g_12)
pyplot.colorbar()
pyplot.title("g_12")

pyplot.subplot(233)
pyplot.pcolor(R, Z, g_13)
pyplot.colorbar()
pyplot.title("g_13")

pyplot.subplot(234)
pyplot.pcolor(R, Z, g_22)
pyplot.colorbar()
pyplot.title("g_22")

pyplot.subplot(235)
pyplot.pcolor(R, Z, g_23)
pyplot.colorbar()
pyplot.title("g_23")

pyplot.subplot(236)
pyplot.pcolor(R, Z, g_33)
pyplot.colorbar()
pyplot.title("g_33")

pyplot.show()

exit(0)



#!/usr/bin/env python3

from boutdata.collect import collect

import matplotlib.pyplot as plt

from numpy import argmin, amax, amin, arange, ndarray


def plot_eigenvals(eigenvalues, eigenvectors=None):
    """Plot the eigenvalues and, optionally, the eigenvectors for a 1D simulation

    eigenvalues should be a 1D array where the odd indices are the
    real components, evens are the imaginary

    eigenvectors should be a 2D array whose first dimension is the
    same length as eigenvalues, with the same interpretation

    """

    if eigenvalues.shape[0] % 2 != 0:
        raise ValueError("Odd number of elements in eigenvalues")

    if eigenvectors is not None:
        # Check that the eigenvectors has the right size
        if len(eigenvectors.shape) != 2:
            raise ValueError("Expecting eigenvectors to be 2D")
        if eigenvectors.shape[0] != len(eigenvalues):
            raise ValueError(
                "First dimension of eigenvectors must match length of eigenvalues"
            )

    # If no eigenvectors supplied, only plot eigenvalues, otherwise
    # eigenvalues and eigenvectors
    nrows = 1 if eigenvectors is None else 2
    fig, ax = plt.subplots(nrows=nrows)

    # If we've only made one plot, ax won't be a list
    if not isinstance(ax, (list, ndarray)):
        ax = [ax]

    eigs_r = eigenvalues[:-1:2]
    eigs_i = eigenvalues[1::2]

    range_r = amax(eigs_r) - amin(eigs_r)
    range_i = amax(eigs_i) - amin(eigs_i)

    ax[0].plot(eigs_r, eigs_i, "x")
    ax[0].set_xlabel("Real component")
    ax[0].set_ylabel("Imaginary component")
    ax[0].set_title("Eigenvalue")

    (overplot,) = ax[0].plot([], [], "ok")

    if eigenvectors is not None:
        # Add a eigenvectors plot
        (vector_r,) = ax[1].plot([], [], "-k", label="Real")
        (vector_i,) = ax[1].plot([], [], "-r", label="Imag")
        ax[1].legend(loc="upper right")
        ax[1].set_xlabel("X")
        ax[1].set_ylabel("Amplitude")
        ax[1].set_title("Eigenvector")
        plt.subplots_adjust(hspace=0.5)

    def onclick(event):
        # Check if user clicked inside the plot
        if event.xdata is None:
            return

        # Find closest eigenvectors point, but stretch axes so
        # real and imaginary components are weighted equally
        if range_r == 0:
            dist = ((eigs_i - event.ydata) / range_i) ** 2
        elif range_i == 0:
            dist = ((eigs_r - event.xdata) / range_r) ** 2
        else:
            dist = ((eigs_r - event.xdata) / range_r) ** 2 + (
                (eigs_i - event.ydata) / range_i
            ) ** 2

        ind = argmin(dist)

        # Update the highlight plot
        overplot.set_data([eigs_r[ind]], [eigs_i[ind]])

        print("Eigenvalue number: %d (%e,%e)" % (ind, eigs_r[ind], eigs_i[ind]))

        if eigenvectors is not None:
            # Update plots
            nx = eigenvectors.shape[1]
            vector_r.set_data(arange(nx), eigenvectors[2 * ind, :])
            vector_i.set_data(arange(nx), eigenvectors[2 * ind + 1, :])
            ax[1].relim()
            ax[1].autoscale_view()

        fig.canvas.draw()

    fig.canvas.mpl_connect("button_press_event", onclick)
    plt.show()


if __name__ == "__main__":
    path = "data"
    eigenvalues = collect("t_array", path=path, info=False)
    eigenvectors = collect("f", xguards=False, path=path, info=False)
    plot_eigenvals(eigenvalues, eigenvectors[..., 0, 0])



#!/usr/bin/env python3

import argparse
import copy
import difflib
import pathlib
import re
import textwrap
import warnings

PHYSICS_MODEL_INCLUDE = '#include "bout/physicsmodel.hxx"'

PHYSICS_MODEL_SKELETON = """
class {name} : public PhysicsModel {{
protected:
  {methods}
}};
"""

PHYSICS_MODEL_RHS_SKELETON = "int {function}({arguments}){override};"

BOUTMAIN = "\n\nBOUTMAIN({})\n"

# Regular expression for a PhysicsModel
PHYSICS_MODEL_RE = re.compile(
    r"""class\s+(?P<name>[a-zA-Z0-9_]+)\s*:   # Class name
    \s*(?:public)?\s*PhysicsModel[\n\s]*{     # Inherits from PhysicsModel
    """,
    re.VERBOSE | re.MULTILINE,
)

FUNCTION_SIGNATURE_ARGUMENT_RE = r"""({arg_type}
    \s+                            # Require spaces only if the argument is named
    (?P<unused{arg_num}>UNUSED\()? # Possible UNUSED macro
    [a-zA-Z_0-9]*                  # Argument name
    (?(unused{arg_num})\))         # If UNUSED macro was present, we need an extra closing bracket
    )?
"""


def create_function_signature_re(function_name, argument_types):
    """Create a regular expression for a legacy physics model function"""

    if not isinstance(argument_types, list):
        argument_types = [argument_types]

    arguments = r",\s*".join(
        [
            FUNCTION_SIGNATURE_ARGUMENT_RE.format(arg_type=argument, arg_num=num)
            for num, argument in enumerate(argument_types)
        ]
    )

    return rf"int\s+{function_name}\s*\({arguments}\)"


LEGACY_MODEL_INCLUDE_RE = re.compile(
    r'^#\s*include.*(<|")boutmain.hxx(>|")', re.MULTILINE
)

BOUT_SOLVE_RE = re.compile(
    r"bout_solve\(([^,)]+,\s*[^,)]+(,\s*[^,)]+)?)\)", re.MULTILINE
)

RHS_RE = re.compile(r"solver\s*->\s*setRHS\(\s*([a-zA-Z0-9_]+)\s*\)")

PRECON_RE = re.compile(r"solver\s*->\s*setPrecon\(\s*([a-zA-Z0-9_]+)\s*\)")

JACOBIAN_RE = re.compile(r"solver\s*->\s*setJacobian\(\s*([a-zA-Z0-9_]+)\s*\)")

SPLIT_OPERATOR_RE = re.compile(
    r"solver\s*->\s*setSplitOperator\(\s*([a-zA-Z0-9_]+),\s*([a-zA-Z0-9_]+)\s*\)"
)


def has_split_operator(source):
    """Return the names of the split operator functions if set, otherwise False"""

    match = SPLIT_OPERATOR_RE.search(source)
    if not match:
        return False

    return match.group(1), match.group(2)


def is_legacy_model(source):
    """Return true if the source is a legacy physics model"""
    return LEGACY_MODEL_INCLUDE_RE.search(source) is not None


def find_last_include(source_lines):
    """Return the line number after the last #include (or 0 if no includes)"""
    for number, line in enumerate(reversed(source_lines)):
        if line.startswith("#include"):
            return len(source_lines) - number
    return 0


def fix_model_operator(
    source, model_name, operator_name, operator_type, new_name, override
):
    """Fix any definitions of the operator, and return the new declaration

    May modify source

    Parameters
    ----------
    source: str
        Source code to fix
    model_name: str
        Name of the PhysicsModel class to create or add methods to
    operator_name: str
        Name of the free function to fix
    operator_type: str, [str]
        Function argument types
    new_name: str
        Name of the PhysicsModel method
    override: bool
        Is `new_name` overriding a virtual?
    """

    # Make sure we have a list of types
    if not isinstance(operator_type, list):
        operator_type = [operator_type]

    # Get a regex for the function signature
    operator_re = re.compile(
        create_function_signature_re(operator_name, operator_type),
        re.VERBOSE | re.MULTILINE,
    )

    # Find any declarations of the free function
    matches = list(operator_re.finditer(source))
    if matches == []:
        warnings.warn(
            f"Could not find {operator_name}; is it defined in another file? If so, you will need to fix it manually"
        )
        return source, False

    # If we found more than one, remove the first one as it's probably
    # a declaration and not a definition
    if len(matches) > 1:
        source = re.sub(
            create_function_signature_re(operator_name, operator_type) + r"\s*;",
            "",
            source,
            flags=re.VERBOSE | re.MULTILINE,
        )

        # Get the names of the function arguments. Every other group
        # from the regex, as the other groups match the `UNUSED` macro
        arg_names = operator_re.search(source).groups()[::2]
    else:
        arg_names = matches[0].groups()[::2]

    # Fix definition and any existing declarations
    arguments = ", ".join(arg_names)

    # Modify the definition: it's out-of-line so we need the qualified name
    modified = operator_re.sub(rf"int {model_name}::{new_name}({arguments})", source)

    # Create the declaration
    return (
        modified,
        PHYSICS_MODEL_RHS_SKELETON.format(
            function=new_name,
            arguments=arguments,
            override=" override" if override else "",
        ),
    )


def fix_bout_constrain(source, error_on_warning):
    """Fix uses of bout_constrain. This is complicated because it might be
    in a conditional, and Solver::constraint returns void

    """

    if "bout_constrain" not in source:
        return source

    # The bout_constrain free function returns False if the Solver
    # doesn't have constraints, but the Solver::constraint method does
    # the checking itself, so we don't need to repeat it
    modified = re.sub(
        r"""if\s*\(\s*(?:!|not)\s*                   # in a conditional, checking for false
            bout_constrain\(([^;]+,[^;]+,[^;]+)\)        # actual function call
            \s*\)                                        # end of conditional
            (?P<brace>\s*{\s*)?                          # possible open brace
            (?:\s*\n)?                                   # possible newline
            \s*throw\s+BoutException\(.*\);(?:\s*\n)?    # throwing an exception
            (?(brace)\s*})?                              # consume matching closing brace
            """,
        r"solver->constraint(\1);\n",
        source,
        flags=re.VERBOSE | re.MULTILINE,
    )

    # The above might not fix everything, so best check if there are any uses left
    remaining_matches = list(re.finditer("bout_constrain", modified))
    if remaining_matches == []:
        # We fixed everything!
        return modified

    # Construct a useful error message
    source_lines = source.splitlines()
    lines_context = []
    for match in remaining_matches:
        bad_line = source[: match.end()].count("\n")
        line_range = range(max(0, bad_line - 1), min(len(source_lines), bad_line + 2))
        lines_context.append(
            "\n  ".join(["{}:{}".format(i, source_lines[i]) for i in line_range])
        )

    message = textwrap.dedent("""\
        Some uses of `bout_constrain` remain, but we could not automatically
        convert them to use `Solver::constraint`. Please fix them before
        continuing:
        """)
    message += "  " + "\n  ".join(lines_context)

    if error_on_warning:
        raise RuntimeError(message)
    print(message)
    return modified


def convert_old_solver_api(source, name):
    """Fix or remove old Solver API calls

    Parameters
    ----------
    source: str
        The source code to modify
    name: str
        The PhysicsModel class name
    """

    # Fixing `bout_solve` is a straight replacement, easy
    source = BOUT_SOLVE_RE.sub(r"solver->add(\1)", source)

    # Completely remove calls to Solver::setRHS
    source = RHS_RE.sub("", source)

    # List of old free functions that now need declarations inside the
    # class definition
    method_decls = []

    # Fix uses of solver->setPrecon
    # Get the name of any free functions passed as arguments to setPrecon
    precons = PRECON_RE.findall(source)
    for precon in precons:
        source, decl = fix_model_operator(
            source,
            name,
            precon,
            ["BoutReal", "BoutReal", "BoutReal"],
            precon,
            override=False,
        )
        if decl:
            method_decls.append(decl)
    # Almost a straight replacement, but it's now a member-function pointer
    source = PRECON_RE.sub(rf"setPrecon(&{name}::\1)", source)

    # Fix uses of solver->setJacobian, basically the same as for setPrecon
    jacobians = JACOBIAN_RE.findall(source)
    for jacobian in jacobians:
        source, decl = fix_model_operator(
            source, name, jacobian, "BoutReal", jacobian, override=False
        )
        if decl:
            method_decls.append(decl)
    source = JACOBIAN_RE.sub(rf"setJacobian(&{name}::\1)", source)

    # If we didn't find any free functions that need to be made into
    # methods, we're done
    if not method_decls:
        return source

    # We need to find the class defintion
    class_def = PHYSICS_MODEL_RE.search(source)
    if class_def is None:
        warnings.warn(
            f"Could not find the '{name}' class to add"
            "preconditioner and/or Jacobian declarations; is it defined"
            "in another file? If so, you will need to fix it manually"
        )
        return source, False

    # The easiest place to stick the method declaration is on the line
    # immediately following the open brace of the class def, and the
    # easiest way to insert it is to split the source into a list,
    # insert in the list, then join the list back into a string.
    # The regex from above finds the offset in the source which we
    # need to turn into a line number
    first_line_of_class = source[: class_def.end() + 1].count("\n")
    methods = "\n  ".join(method_decls)
    source_lines = source.splitlines()
    source_lines.insert(first_line_of_class, f"  {methods}")

    return "\n".join(source_lines)


def convert_legacy_model(source, name, error_on_warning):
    """Convert a legacy physics model to a PhysicsModel"""

    if not is_legacy_model(source):
        return source

    source = fix_bout_constrain(source, error_on_warning)

    # Replace legacy header
    source = LEGACY_MODEL_INCLUDE_RE.sub(r"#include \1bout/physicsmodel.hxx\2", source)

    method_decls = []

    source, decl = fix_model_operator(
        source, name, "physics_init", "bool", "init", override=True
    )
    if decl:
        method_decls.append(decl)

    split_operators = has_split_operator(source)
    if split_operators:
        source = SPLIT_OPERATOR_RE.sub(r"setSplitOperator(true)", source)

        convective, diffusive = split_operators
        # Fix the free functions
        source, decl = fix_model_operator(
            source, name, convective, "BoutReal", "convective", override=True
        )
        if decl:
            method_decls.append(decl)
        source, decl = fix_model_operator(
            source, name, diffusive, "BoutReal", "diffusive", override=True
        )
        if decl:
            method_decls.append(decl)
    else:
        # Fix the rhs free function
        source, decl = fix_model_operator(
            source, name, "physics_run", "BoutReal", "rhs", override=True
        )
        if decl:
            method_decls.append(decl)

    source_lines = source.splitlines()
    last_include = find_last_include(source_lines)

    methods = "\n  ".join(method_decls)
    physics_model = PHYSICS_MODEL_SKELETON.format(methods=methods, name=name)

    source_lines.insert(last_include, physics_model)
    source_lines.append(BOUTMAIN.format(name))

    return "\n".join(source_lines)


def yes_or_no(question):
    """Convert user input from yes/no variations to True/False"""
    while True:
        reply = input(question + " [y/N] ").lower().strip()
        if not reply or reply[0] == "n":
            return False
        if reply[0] == "y":
            return True


def create_patch(filename, original, modified):
    """Create a unified diff between original and modified"""

    patch = "\n".join(
        difflib.unified_diff(
            original.splitlines(),
            modified.splitlines(),
            fromfile=filename,
            tofile=filename,
            lineterm="",
        )
    )

    return patch


if __name__ == "__main__":
    parser = argparse.ArgumentParser(
        formatter_class=argparse.RawDescriptionHelpFormatter,
        description=textwrap.dedent("""\
            Upgrade legacy physics models to use the PhysicsModel class

            This will do the bare minimum required to compile, and
            won't make global objects (like Field3Ds) members of the
            new class, or free functions (other than
            `physics_init`/`physics_run`, preconditioners, and
            Jacobians) methods of the new class. Comments may also be
            left behind.

            By default, this will use the file name stripped of file
            extensions as the name of the new class. Use '--name=<new
            name>' to give a different name.
            """),
    )

    parser.add_argument("files", action="store", nargs="+", help="Files to fix")
    parser.add_argument(
        "--force", "-f", action="store_true", help="Make changes without asking"
    )
    parser.add_argument(
        "--quiet", "-q", action="store_true", help="Don't print patches"
    )
    parser.add_argument(
        "--patch-only", "-p", action="store_true", help="Print the patches and exit"
    )
    parser.add_argument(
        "--name",
        "-n",
        action="store",
        nargs="?",
        type=str,
        help="Name for new PhysicsModel class, default is from filename",
    )

    args = parser.parse_args()

    if args.force and args.patch_only:
        raise ValueError("Incompatible options: --force and --patch")

    for filename in args.files:
        with open(filename, "r") as f:
            contents = f.read()

        original = copy.deepcopy(contents)

        match = PHYSICS_MODEL_RE.search(original)
        if match is not None:
            new_name = match.group("name")
        else:
            new_name = args.name or pathlib.Path(filename).stem.capitalize().replace(
                " ", "_"
            )

        try:
            if re.match(r"^[0-9]+.*", new_name) and not args.force:
                raise ValueError(
                    f"Invalid name: '{new_name}'. Use --name to specify a valid C++ identifier"
                )

            modified = convert_legacy_model(
                original, new_name, not (args.force or args.patch_only)
            )

            modified = convert_old_solver_api(modified, new_name)
        except (RuntimeError, ValueError) as e:
            error_message = textwrap.indent(f"{e}", " ")
            print(
                f"There was a problem applying automatic fixes to {filename}:\n\n{error_message}"
            )
            continue

        patch = create_patch(filename, original, modified)

        if args.patch_only:
            print(patch)
            continue

        if not patch:
            if not args.quiet:
                print("No changes to make to {}".format(filename))
            continue

        if not args.quiet:
            print("\n******************************************")
            print("Changes to {}\n".format(filename))
            print(patch)
            print("\n******************************************")

        make_change = args.force or yes_or_no("Make changes to {}?".format(filename))

        if make_change:
            with open(filename, "w") as f:
                f.write(modified)



#!/usr/bin/env python

#
# Generate an input mesh
#

from boututils.datafile import DataFile  # Wrapper around NetCDF4 libraries

nx = 5  # Minimum is 5: 2 boundary, one evolved
ny = 32  # Minimum 5. Should be divisible by number of processors (so powers of 2 nice)
dy = 1.0  # distance between points in y, in m/g22/lengthunit
ixseps1 = -1
ixseps2 = -1

f = DataFile()
f.open("test-staggered.nc", create=True)

f.write("nx", nx)
f.write("ny", ny)
f.write("dy", dy)
f.write("ixseps1", ixseps1)
f.write("ixseps2", ixseps2)

f.close()



#!/usr/bin/env python3
from boututils.datafile import DataFile
import matplotlib.pyplot as plt

path = "data"
monitors = [
    ["PROBES", "T_up"],
    ["PROBES", "n_up"],
    ["slow", "T_corner"],
]

for pack in monitors:
    filename, data_name = pack
    t = DataFile(path + "/" + filename + ".dmp.0.nc").read("t_array")
    data = DataFile(path + "/" + filename + ".dmp.0.nc").read(data_name).flatten()
    plt.plot(t, data, label="{} {}".format(filename, data_name))

time = DataFile(path + "/BOUT.dmp.0.nc").read("t_array")
data = DataFile(path + "/BOUT.dmp.0.nc").read("T")[:, 2, 2, 0]

plt.plot(time, data, marker="+", label="BOUT++ T")

plt.xlabel("Time")
plt.legend()
plt.show()



from __future__ import print_function
####################################
# Replicates IDL's showdata and more
####################################

import numpy as np

from tvtk.tools import visual

try:
    from enthought.mayavi import mlab
except ImportError:
    try:
        from mayavi import mlab
    except ImportError:
        print("No mlab available")

from boututils.anim import anim
###########################
# Read polslice array

npzfile = np.load("pslice.npz")
r = npzfile["r"]
z = npzfile["z"]
fm = npzfile["fm"]

########################################################
# Set up the window

f = mlab.figure(size=(800, 600))
# Tell visual to use this as the viewer.
visual.set_viewer(f)

########################################################
# Do the appropriate graph

# s = mlab.contour_surf(r,z,fun, contours=30, line_width=.5, transparent=True)
# s=mlab.surf(r,z,fun, colormap='Spectral')
s = mlab.mesh(
    r, z, fm[0, :, :], scalars=fm[0, :, :], colormap="PuOr"
)  # , wrap_scale='true')#, representation='wireframe')
s.enable_contours = True
s.contour.filled_contours = True

# Define perspective and optional attributes. You can also implement from the window afterwards

mlab.view(0, 0)
# mlab.view(-94.159958841373324,
# 53.777002382688906,
# 8.2311808018087582)
mlab.draw(f)
mlab.colorbar(orientation="vertical")
# mlab.axes()
# mlab.outline()


########################################################
# mlab animation

anim(s, fm, save=True)



# -*- coding: utf-8 -*-
# Replicate the graphs from Xu_3fields_hands-on.pdf (BOUT++ workshop 2013)
# needs runs in data & data0 folders
# After running the cases define path0 and path1 accordingly
from __future__ import division
from past.utils import old_div
import numpy as np
from boututils.datafile import DataFile
from boutdata.collect import collect
from pylab import plot, show, figure, xlabel, ylabel, annotate, xlim, ylim
from boututils.moment_xyzt import moment_xyzt
from boututils.mode_structure import mode_structure
from boututils.plotpolslice import plotpolslice
from boututils.calculus import deriv
from mayavi import mlab


path0 = "./data0/"
path1 = "./data/"

period = 15

gfile = "./cbm18_dens8.grid_nx68ny64.nc"


with DataFile(gfile) as f:
    g = {v: f.read(v) for v in f.keys()}


Dphi0 = collect("Dphi0", path=path0)
phi0 = collect("phi0", path=path1)  # needs diamagnetic effects
#
psixy = g.get("psixy")
PSI_AXIS = g.get("psi_axis")
PSI_BNDRY = g.get("psi_bndry")
#
psix = old_div((psixy[:, 32] - PSI_AXIS), (PSI_BNDRY - PSI_AXIS))
Epsi = -deriv(phi0[:, 32], psix)
#
#
fig = figure()
plot(psix, -Dphi0[:, 32], "r", linewidth=5)
plot(psix, Epsi, "k", linewidth=5)
annotate(
    "w/o flow",
    xy=(0.3, 0.7),
    xycoords="axes fraction",
    horizontalalignment="center",
    verticalalignment="center",
    size=30,
)
annotate(
    "w/ flow",
    xy=(0.7, 0.4),
    xycoords="axes fraction",
    horizontalalignment="center",
    verticalalignment="center",
    color="r",
    size=30,
)
xlabel("Radial $\psi$", fontsize=25)
ylabel("$\Omega(\psi)/\omega_A$", fontsize=25)
ylim([-0.05, 0])
xlim([0.4, 1.2])
fig.set_tight_layout(True)
show(block=False)

p_f0 = collect("P", path=path0)
p_f = collect("P", path=path1)
#
rmsp_f0 = moment_xyzt(p_f0, "RMS").rms
rmsp_f = moment_xyzt(p_f, "RMS").rms
#

fig = figure(figsize=(10, 8))
plot(np.gradient(np.log(rmsp_f0[:, 34, 32])), color="k", linewidth=3)
plot(np.gradient(np.log(rmsp_f[:, 34, 32])), color="red", linewidth=3)

ylabel("$\gamma / \omega_A$", fontsize=25)
xlabel("Time$(\\tau_A)$", fontsize=25)
annotate(
    "w/o flow",
    xy=(0.5, 0.7),
    xycoords="axes fraction",
    horizontalalignment="center",
    verticalalignment="center",
    size=30,
)
annotate(
    "w/ flow",
    xy=(0.5, 0.4),
    xycoords="axes fraction",
    horizontalalignment="center",
    verticalalignment="center",
    color="r",
    size=30,
)
ylim([0, 0.5])
xlim([0, 100])
fig.set_tight_layout(True)
show(block=False)


plotpolslice(p_f0[50, :, :, :], gfile, period=period, fig=1)
mlab.text(0.01, 0.99, "w/o flow")

plotpolslice(p_f[50, :, :, :], gfile, period=period, fig=1)
mlab.text(0.01, 0.99, "w/ flow")

fig = figure()
mode_structure(p_f0[50, :, :, :], g, period=period)
plot([40, 40], [0, 0.014], "k--", linewidth=5)
annotate(
    "w/o flow",
    xy=(0.3, 0.7),
    xycoords="axes fraction",
    horizontalalignment="center",
    verticalalignment="center",
    size=30,
)
ylim([0, 0.014])
xlim([0, 80])
fig.set_tight_layout(True)
show(block=False)


figure()
mode_structure(p_f[50, :, :, :], g, period=period)
plot([40, 40], [0, 0.014], "k--", linewidth=5)
annotate(
    "w/ flow",
    xy=(0.3, 0.7),
    xycoords="axes fraction",
    horizontalalignment="center",
    verticalalignment="center",
    color="k",
    size=30,
)
ylim([0, 0.0001])
xlim([0, 80])
show(block=False)
show()



from __future__ import print_function
from __future__ import division
from past.utils import old_div

# Generate MMS solutions for a tokamak geometry

from boutdata.mms import *

from sympy import sin, cos

shape = SimpleTokamak()
metric = shape.metric()  # Get the metric tensor

###
# Define solution in normalised x,y coordinates
# NOTE: These are orthogonal tokamak coordinates
#       so y is poloidal angle, not parallel coordinate

drive = sin(6 * x**2 - z + y)  # Constant drive for advection

advect = cos(4 * x**2 + z - y) + sin(t) * sin(3 * x + 2 * z - y)
delp2 = cos(4 * x**2 + z - y) + sin(t) * sin(3 * x + 2 * z - y)
laplacepar = cos(4 * x**2 + z - y) + sin(t) * sin(3 * x + 2 * z - y)

ZMAX = 1

# Turn solution into real x and z coordinates
# NOTE: Z is shifted, so y is now the parallel coordinate
zShift = shape.zShift.subs(x, old_div(metric.x, metric.Lx))
sinty = shape.sinty.subs(x, old_div(metric.x, metric.Lx))
replace = [(x, old_div(metric.x, metric.Lx)), (z, old_div(metric.z, ZMAX) + zShift)]

drive = drive.subs(replace)
advect = advect.subs(replace)
delp2 = delp2.subs(replace)
laplacepar = laplacepar.subs(replace)

# Calculate time derivatives

dadt = -1e-3 * bracket(drive, advect, metric)

dddt = 1e-5 * Delp2(delp2, metric)
dgdt = Laplace_par(laplacepar, metric)

vars = [
    (advect, dadt, "advect"),
    (delp2, dddt, "delp2"),
    (laplacepar, dgdt, "laplacepar"),
]


# Change back to normalised coordinates
# NOTE: Z remains shifted, so y is still parallel coordinate
#       Need to scale the toroidal angle part of z, not zShift part
replace = [
    (metric.x, x * metric.Lx),
    (metric.z, (z - shape.zShift) * ZMAX + shape.zShift),
]

drive = drive.subs(replace)
print("[drive]")
print("solution = " + exprToStr(drive))

# Loop over variables and print solution, source etc.
for f, dfdt, name in vars:
    # Calculate source
    S = diff(f, t) - dfdt

    # Differentials for boundary conditions
    dfdx = diff(f, metric.x)
    dfdy = diff(f, metric.y)

    # Substitute back to get in terms of x,y,z

    f = f.subs(replace)
    dfdt = dfdt.subs(replace)
    S = S.subs(replace)
    dfdx = dfdx.subs(replace)
    dfdy = dfdy.subs(replace)

    print("\n[" + name + "]")
    print("solution = " + exprToStr(f))
    print("\nddx = " + exprToStr(dfdx))
    print("\nddy = " + exprToStr(dfdy))
    print("\nsource = " + exprToStr(S))

    # print("\nSource magnitude: %e" % exprMag(S))
